__title__ = "sewer"
__description__ = "Sewer is a programmatic Lets Encrypt(ACME) client"
__url__ = "https://github.com/komuw/sewer"
__version__ = "0.8.2"
__author__ = "komuW"
__author_email__ = "komuw05@gmail.com"
__license__ = "MIT"



### legacy DNS bindings - DEPRECATED ###

from .dns_providers import AuroraDns  # noqa: F401
from .dns_providers import CloudFlareDns  # noqa: F401
from .dns_providers import AcmeDnsDns  # noqa: F401
from .dns_providers import AliyunDns  # noqa:F401
from .dns_providers import HurricaneDns  # noqa:F401
from .dns_providers import RackspaceDns  # noqa:F401
from .dns_providers import DNSPodDns  # noqa:F401
from .dns_providers import DuckDNSDns  # noqa:F401
from .dns_providers import ClouDNSDns  # noqa:F401
from .dns_providers import Route53Dns  # noqa:F401
from .dns_providers import PowerDNSDns  # noqa:F401
from .dns_providers import GandiDns  # noqa:F401



import json

try:
    aliyun_dependencies = True
    from aliyunsdkcore import client
    from aliyunsdkalidns.request.v20150109 import DescribeDomainRecordsRequest
    from aliyunsdkalidns.request.v20150109 import AddDomainRecordRequest
    from aliyunsdkalidns.request.v20150109 import DeleteDomainRecordRequest
except ImportError:
    aliyun_dependencies = False

from . import common


class _ResponseForAliyun(object):
    """
    wrapper aliyun resp to the format sewer wanted.
    """

    def __init__(self, status_code=200, content=None, headers=None):
        self.status_code = status_code
        self.headers = headers or {}
        self.content = content or {}
        self.content = json.dumps(content)
        super(_ResponseForAliyun, self).__init__()

    def json(self):
        return json.loads(self.content)


class AliyunDns(common.BaseDns):
    def __init__(self, key, secret, endpoint="cn-beijing", debug=False):
        """
        aliyun dns client
        :param str key: access key
        :param str secret: access sceret
        :param str endpoint: endpoint
        :param bool debug: if debug?
        """
        super(AliyunDns, self).__init__()
        if not aliyun_dependencies:
            raise ImportError(
                """You need to install aliyunDns dependencies. run; pip3 install sewer[aliyun]"""
            )
        self._key = key
        self._secret = secret
        self._endpoint = endpoint
        self._debug = debug
        self.clt = client.AcsClient(self._key, self._secret, self._endpoint, debug=self._debug)

    def _send_reqeust(self, request):
        """
        send request to aliyun
        """
        request.set_accept_format("json")
        try:
            status, headers, result = self.clt.implementation_of_do_action(request)
            if isinstance(result, bytes):
                result = result.decode()
            result = json.loads(result)
            if "Message" in result or "Code" in result:
                result["Success"] = False
                self.logger.warning("aliyundns resp error: %s", result)
        except Exception as exc:
            self.logger.warning("aliyundns failed to send request: %s, %s", str(exc), request)
            status, headers, result = 502, {}, b'{"Success": false}'
            result = json.loads(result)

        if self._debug:
            self.logger.info("aliyundns request name: %s", request.__class__.__name__)
            self.logger.info("aliyundns request query: %s", request.get_query_params())
        return _ResponseForAliyun(status, result, headers)

    def query_recored_items(self, host, zone=None, tipe=None, page=1, psize=200):
        """
        query recored items.
        :param str host: like example.com
        :param str zone: like menduo.example.com
        :param str tipe: TXT, CNAME, IP or other
        :param int page:
        :param int psize:
        :return dict: res = {
                'DomainRecords':
                    {'Record': [
                        {
                            'DomainName': 'menduo.net',
                            'Line': 'default',
                            'Locked': False,
                            'RR': 'zb',
                            'RecordId': '3989515483698964',
                            'Status': 'ENABLE',
                            'TTL': 600,
                            'Type': 'A',
                            'Value': '127.0.0.1',
                            'Weight': 1
                        },
                        {
                            'DomainName': 'menduo.net',
                            'Line': 'default',
                            'Locked': False,
                            'RR': 'a.sub',
                            'RecordId': '3989515480778964',
                            'Status': 'ENABLE',
                            'TTL': 600,
                            'Type': 'CNAME',
                            'Value': 'h.p.menduo.net',
                            'Weight': 1
                        }
                    ]
                    },
                'PageNumber': 1,
                'PageSize': 20,
                'RequestId': 'FC4D02CD-EDCC-4EE8-942F-1497CCC3B10E',
                'TotalCount': 95
            }
        """
        request = DescribeDomainRecordsRequest.DescribeDomainRecordsRequest()
        request.get_action_name()
        request.set_DomainName(host)
        request.set_PageNumber(page)
        request.set_PageSize(psize)
        if zone:
            request.set_RRKeyWord(zone)
        if tipe:
            request.set_TypeKeyWord(tipe)
        resp = self._send_reqeust(request)
        body = resp.json()
        return body

    def query_recored_id(self, root, zone, tipe="TXT"):
        """
        find recored
        :param str root: root host, like example.com
        :param str zone: sub zone, like menduo.example.com
        :param str tipe: record tipe, TXT, CNAME, IP. we use TXT
        :return str:
        """
        record_id = None
        recoreds = self.query_recored_items(root, zone, tipe=tipe)
        recored_list = recoreds.get("DomainRecords", {}).get("Record", [])
        recored_item_list = [i for i in recored_list if i["RR"] == zone]
        if len(recored_item_list):
            record_id = recored_item_list[0]["RecordId"]
        return record_id

    @staticmethod
    def extract_zone(domain_name):
        """
        extract domain to root, sub, acme_txt
        :param str domain_name: the value sewer client passed in, like *.menduo.example.com
        :return tuple: root, zone, acme_txt
        """
        if domain_name.count(".") > 1:
            zone, middle, last = str(domain_name).rsplit(".", 2)
            root = ".".join([middle, last])
            acme_txt = "_acme-challenge.%s" % zone
        else:
            zone = ""
            root = domain_name
            acme_txt = "_acme-challenge"
        return root, zone, acme_txt

    def create_dns_record(self, domain_name, domain_dns_value):
        """
        create a dns record
        :param str domain_name: the value sewer client passed in, like *.menduo.example.com
        :param str domain_dns_value: the value sewer client passed in.
        :return _ResponseForAliyun:
        """
        self.logger.info("create_dns_record start: %s", (domain_name, domain_dns_value))
        root, _, acme_txt = self.extract_zone(domain_name)

        request = AddDomainRecordRequest.AddDomainRecordRequest()
        request.set_DomainName(root)
        request.set_TTL(600)
        request.set_RR(acme_txt)
        request.set_Type("TXT")
        request.set_Value(domain_dns_value)
        resp = self._send_reqeust(request)

        self.logger.info("create_dns_record end: %s", (domain_name, domain_dns_value, resp.json()))

        return resp

    def delete_dns_record(self, domain_name, domain_dns_value):
        """
        delete a txt record we created just now.
        :param str domain_name: the value sewer client passed in, like *.menduo.example.com
        :param str domain_dns_value: the value sewer client passed in. we do not use this.
        :return _ResponseForAliyun:
        :return:
        """
        self.logger.info("delete_dns_record start: %s", (domain_name, domain_dns_value))

        root, _, acme_txt = self.extract_zone(domain_name)

        record_id = self.query_recored_id(root, acme_txt)
        if not record_id:
            msg = "failed to find record_id of domain: %s, value: %s", domain_name, domain_dns_value
            self.logger.warning(msg)
            return

        self.logger.info("start to delete dns record, id: %s", record_id)

        request = DeleteDomainRecordRequest.DeleteDomainRecordRequest()
        request.set_RecordId(record_id)
        resp = self._send_reqeust(request)

        self.logger.info("delete_dns_record end: %s", (domain_name, domain_dns_value, resp.json()))
        return resp



### legacy DNS bindings - DEPRECATED ###

from .auroradns import AuroraDns  # noqa: F401
from .cloudflare import CloudFlareDns  # noqa: F401
from .acmedns import AcmeDnsDns  # noqa: F401
from .aliyundns import AliyunDns  # noqa: F401
from .hurricane import HurricaneDns  # noqa: F401
from .rackspace import RackspaceDns  # noqa: F401
from .dnspod import DNSPodDns  # noqa: F401
from .duckdns import DuckDNSDns  # noqa: F401
from .cloudns import ClouDNSDns  # noqa: F401
from .route53 import Route53Dns  # noqa: F401
from .powerdns import PowerDNSDns
from .gandi import GandiDns



try:
    cloudns_dependencies = True
    from cloudns_api import record
except ImportError:
    cloudns_dependencies = False

from . import common


### FIX ME ### this assumes there are only two levels above the host (no bbc.co.uk eg.)


def _split_domain_name(domain_name):
    """ClouDNS requires the domain name and host to be split."""
    full_domain_name = "_acme-challenge.{}".format(domain_name)
    domain_parts = full_domain_name.split(".")

    domain_name = ".".join(domain_parts[-2:])
    host = ".".join(domain_parts[:-2])

    return domain_name, host


class ClouDNSDns(common.BaseDns):

    dns_provider_name = "cloudns"

    def __init__(self, *args, **kwargs):

        if not cloudns_dependencies:
            raise ImportError(
                """You need to install ClouDNSDns dependencies. run; pip3 install sewer[cloudns]"""
            )

        super(ClouDNSDns, self).__init__(*args, **kwargs)

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record")
        domain_name, host = _split_domain_name(domain_name)
        response = record.create(
            domain_name=domain_name, host=host, record_type="TXT", record=domain_dns_value, ttl=60
        )

        if not response.success:
            self.logger.info("ClouDNS could not create DNS record.")
            raise Exception("ClouDNS responded with an error.")

        self.logger.info("create_dns_record_success")
        return

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record")
        domain_name, host = _split_domain_name(domain_name)
        response = record.list(domain_name=domain_name, host=host, record_type="TXT")

        if not response.success:
            self.logger.info("ClouDNS could not find DNS record to delete.")
            raise Exception("ClouDNS responded with an error.")

        for record_id, item in response.payload.items():
            if item["record"] == domain_dns_value:
                response = record.delete(domain_name=domain_name, record_id=record_id)
                if not response.success:
                    self.logger.info("ClouDNS could not delete DNS record.")
                    raise Exception("ClouDNS responded with an error.")
                self.logger.info("delete_dns_record_success")
                return

        self.logger.info("ClouDNS could not find DNS record to delete.")
        raise Exception("ClouDNS responded with an error.")



import base64, logging
from typing import Union


def log_response(response) -> str:
    """
    renders a python-requests response as json or as a string
    """
    try:
        log_body = response.json()
    except ValueError:
        log_body = response.content[:40]
    return log_body


def create_logger(name: str, log_level: Union[str, int]) -> logging.Logger:
    """
    return a logger configured with name and log_level
    """

    logger = logging.getLogger(name)
    logger.setLevel(log_level)
    if not logger.hasHandlers():
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    return logger


def safe_base64(un_encoded_data: Union[str, bytes]) -> str:
    "return ACME-safe base64 encoding of un_encoded_data"

    if isinstance(un_encoded_data, str):
        un_encoded_data = un_encoded_data.encode("utf8")
    r = base64.urlsafe_b64encode(un_encoded_data).rstrip(b"=")
    return r.decode("utf8")



import urllib.parse
import requests

from . import common


class DuckDNSDns(common.BaseDns):

    dns_provider_name = "duckdns"

    def __init__(self, duckdns_token, DUCKDNS_API_BASE_URL="https://www.duckdns.org"):

        self.duckdns_token = duckdns_token
        self.HTTP_TIMEOUT = 65  # seconds

        if DUCKDNS_API_BASE_URL[-1] != "/":
            self.DUCKDNS_API_BASE_URL = DUCKDNS_API_BASE_URL + "/"
        else:
            self.DUCKDNS_API_BASE_URL = DUCKDNS_API_BASE_URL
        super(DuckDNSDns, self).__init__()

    def _common_dns_record(self, logger_info, domain_name, payload_end_arg):
        self.logger.info("{0}".format(logger_info))
        # add provider domain to the domain name if not present
        provider_domain = ".duckdns.org"
        if domain_name.rfind(provider_domain) == -1:
            "".join((domain_name, provider_domain))

        url = urllib.parse.urljoin(self.DUCKDNS_API_BASE_URL, "update")

        payload = dict([("domains", domain_name), ("token", self.duckdns_token), payload_end_arg])
        update_duckdns_dns_record_response = requests.get(
            url, params=payload, timeout=self.HTTP_TIMEOUT
        )

        normalized_response = update_duckdns_dns_record_response.text
        self.logger.debug(
            "update_duckdns_dns_record_response. status_code={0}. response={1}".format(
                update_duckdns_dns_record_response.status_code, normalized_response
            )
        )

        if update_duckdns_dns_record_response.status_code != 200 or normalized_response != "OK":
            # raise error so that we do not continue to make calls to DuckDNS
            # server
            raise ValueError(
                "Error creating DuckDNS dns record: status_code={status_code} response={response}".format(
                    status_code=update_duckdns_dns_record_response.status_code,
                    response=normalized_response,
                )
            )
        self.logger.info("{0}_success".format(logger_info))

    def create_dns_record(self, domain_name, domain_dns_value):
        self._common_dns_record("create_dns_record", domain_name, ("txt", domain_dns_value))

    def delete_dns_record(self, domain_name, domain_dns_value):
        self._common_dns_record("delete_dns_record", domain_name, ("clear", "true"))



import json
import requests

from . import common


class PowerDNSDns(common.BaseDns):
    """
    For PowerDNS, all subdomains for a given domain live under the apex zone `zone_id`.
    For example, if you want a cert for domain.tld and www.domain.tld, you need
    to create two DNS records:
    1) `acme-challenge.domain.tld. IN TXT`
    2) `acme-challenge.www.domain.tld. IN TXT`

    However, both of these records must be created under `/servers/{server_id}/zones/{zone_id}`,
    where `zone_id` is the apex domain (`domain.tld`)

    So, we must be smart about stripping out subdomains as part of the URL passed
    to `requests.patch`, but must maintain the FQDN in the `name` field of the `payload`.
    """

    dns_provider_name = "powerdns"

    def __init__(self, powerdns_api_key, powerdns_api_url):
        self.powerdns_api_key = powerdns_api_key
        self.powerdns_api_url = powerdns_api_url
        super(PowerDNSDns, self).__init__()

    def validate_powerdns_zone(self, domain_name):
        """
        Walk `domain_name` backwards, trying to find the apex domain.
        E.g.: For `fu.bar.baz.domain.com`, `response.status_code` will only be
        `200` for `domain.com`
        """
        d = "."
        count = domain_name.count(d)

        while True:
            url = self.powerdns_api_url + "/" + domain_name
            response = requests.get(url, headers={"X-API-Key": self.powerdns_api_key})

            if response.status_code == 200:
                return domain_name
            elif count <= 0:
                raise ValueError(
                    "Could not determine apex domain: (count: %s, domain_name: %s)"
                    % (count, domain_name)
                )
            else:
                split = domain_name.split(d)
                split.pop(0)
                domain_name = d.join(split)
                count -= 1

    def _common_dns_record(self, domain_name, domain_dns_value, changetype):
        if changetype not in ("REPLACE", "DELETE"):
            raise ValueError("changetype is not valid.")

        payload = {
            "rrsets": [
                {
                    "name": "_acme-challenge" + "." + domain_name + ".",
                    "type": "TXT",
                    "ttl": 60,
                    "changetype": changetype,
                    "records": [{"content": f'"{domain_dns_value}"', "disabled": False}],
                }
            ]
        }
        self.logger.debug("PowerDNS domain name: %s", domain_name)
        self.logger.debug("PowerDNS payload: %s", payload)

        apex_domain = self.validate_powerdns_zone(domain_name)
        url = self.powerdns_api_url + "/" + apex_domain
        self.logger.debug("apex_domain: %s", apex_domain)
        self.logger.debug("url: %s", url)

        try:
            response = requests.patch(
                url, data=json.dumps(payload), headers={"X-API-Key": self.powerdns_api_key}
            )
            self.logger.debug("PowerDNS response: %s, %s", response.status_code, response.text)
        except requests.exceptions.RequestException as e:
            self.logger.error("Unable to communicate with PowerDNS API: %s", e)
            raise

        # Per https://doc.powerdns.com/authoritative/http-api/zone.html:
        # PATCH /servers/{server_id}/zones/{zone_id}
        # Creates/modifies/deletes RRsets present in the payload and their comments.
        # Returns 204 No Content on success.
        if response.status_code != 204:
            raise ValueError("Error creating or deleting PowerDNS record: %s" % response.text)

    def create_dns_record(self, domain_name, domain_dns_value):
        self._common_dns_record(domain_name, domain_dns_value, "REPLACE")

    def delete_dns_record(self, domain_name, domain_dns_value):
        self._common_dns_record(domain_name, domain_dns_value, "DELETE")



import os
import requests

from . import common


class GandiDns(common.BaseDns):
    """
    """

    dns_provider_name = "gandi"

    def __init__(
        self,
        GANDI_API_KEY=None,
        GANDI_API_BASE_URL="https://dns.api.gandi.net/api/v5/",
        DEFAULT_TTL=10800,
        requests_lib=requests,
    ):
        self.GANDI_API_KEY = GANDI_API_KEY
        self.HTTP_TIMEOUT = 65  # seconds
        self.RECORD_TTL = DEFAULT_TTL
        self.requests = requests_lib

        if GANDI_API_BASE_URL[-1] != "/":
            self.GANDI_API_BASE_URL = GANDI_API_BASE_URL + "/"
        else:
            self.GANDI_API_BASE_URL = GANDI_API_BASE_URL

        # pass an API key
        if not GANDI_API_KEY:
            raise ValueError("Error initializing Gandi DNS adapter. Pass an API key.")

        self.GET_HEADERS = {"X-Api-Key": self.GANDI_API_KEY}
        self.POST_HEADERS = {"X-Api-Key": self.GANDI_API_KEY, "Content-Type": "application/json"}

        super(GandiDns, self).__init__()

    def create_dns_record(self, domain_name, domain_dns_value):
        self.delete_record(domain_name, idempotent=True)
        [subdomain, base_domain] = GandiDns.split_domain(domain_name)
        zone_records_href = self.get_zone_records_href(base_domain)

        request_body = {
            "rrset_type": "TXT",
            "rrset_ttl": self.RECORD_TTL,
            "rrset_name": GandiDns.subdomain_to_challenge_domain(subdomain),
            "rrset_values": [domain_dns_value],
        }

        create_record_resp = self.requests.post(
            zone_records_href, headers=self.POST_HEADERS, json=request_body
        )
        if not create_record_resp.status_code < 300:
            raise RuntimeError("createRecord failed")

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.delete_record(domain_name)

    def delete_record(self, domain_name, idempotent=False):
        def get_subdomain_record(subdomain, all_records):
            subdomain_records = list(
                filter(lambda rec: rec["rrset_name"] == subdomain, all_records)
            )
            if not len(subdomain_records) == 1:
                if idempotent:
                    return None
                else:
                    raise RuntimeError("there is not exactly one record for domain: " + domain_name)

            return subdomain_records[0]

        [subdomain, base_domain] = GandiDns.split_domain(domain_name)
        zone_records_href = self.get_zone_records_href(base_domain)
        all_records = self.get_all_zone_records(zone_records_href)
        subdomain_record = get_subdomain_record(
            GandiDns.subdomain_to_challenge_domain(subdomain), all_records
        )

        if subdomain_record:
            del_record_resp = self.requests.delete(
                subdomain_record["rrset_href"], headers=self.GET_HEADERS
            )
            if not del_record_resp.status_code < 300:
                raise RuntimeError("deleteRecord failed")

    def get_zone_records_href(self, base_domain):
        domain_resp = self.requests.get(
            os.path.join(self.GANDI_API_BASE_URL, "domains", base_domain), headers=self.GET_HEADERS
        )

        if not domain_resp.status_code < 300:
            raise RuntimeError("getZoneHref failed")

        domain_info = domain_resp.json()
        return domain_info["zone_records_href"]

    def get_all_zone_records(self, zone_records_href):
        all_records_resp = self.requests.get(zone_records_href, headers=self.GET_HEADERS)
        if not all_records_resp.status_code < 300:
            raise RuntimeError("getAllZoneRecords failed")

        all_records = all_records_resp.json()
        return all_records

    @staticmethod
    def split_domain(domain_name):
        def has_second_level_domain(split_domains):
            return len(split_domains) >= 2

        split_domains = domain_name.split(".")
        if not has_second_level_domain(split_domains):
            raise RuntimeError(
                "domain: " + str(domain_name) + "does not have a second level domain"
            )

        separator = "."
        base_domain = separator.join(split_domains[-2:])
        subdomain = separator.join(split_domains[0:-2])
        if not subdomain:
            subdomain = "@"

        return [subdomain, base_domain]

    @staticmethod
    def subdomain_to_challenge_domain(subdomain):
        return "_acme-challenge" + "." + subdomain



from .lib import create_logger
from typing import Any, Dict, Sequence, Tuple, Union

ErrataItemType = Tuple[str, str, Dict[str, str]]


class ProviderBase:
    """
    New-model driver documentation is in docs/UnifiedProvider.md
    """

    def __init__(self, **kwargs: Any) -> None:
        # REQUIRED chal_types - with "fail fast" strict checking
        chal_types = kwargs.pop("chal_types")
        if not isinstance(chal_types, (list, tuple)) or any(
            not isinstance(ct, str) for ct in chal_types
        ):
            raise ValueError("chal_types must be a list or tuple of strings, not: %s" % chal_types)
        self.chal_types = chal_types

        # OPTIONAL logger or LOG_LEVEL (obsolescent) - setup logging; exactly one is accepted!
        if "logger" in kwargs:
            self.logger = kwargs.pop("logger")
        else:
            self.logger = create_logger(__name__, kwargs.pop("LOG_LEVEL", "INFO"))

        # OPTIONAL prop_timeout - how long to wait for unpropagated list to clear?
        self.prop_timeout = kwargs.pop("prop_timeout", 0)

        # OPTIONAL prop_sleep_times - int or sequence of ints of seconds to sleep
        if "prop_sleep_times" not in kwargs:
            self.prop_sleep_times = [1, 2, 4, 8]  # default: sleep for 1, 2, 4, 8, 8 ...
        else:
            pst = kwargs.pop("prop_sleep_times")
            if isinstance(pst, int):
                self.prop_sleep_times = [pst]
            elif isinstance(pst, (list, tuple)):
                if any(not isinstance(value, int) for value in pst):
                    raise ValueError("prop_sleep_times list includes non-int value: %s" % pst)
                self.prop_sleep_times = list(pst)
            else:
                raise ValueError("prop_sleep_time must be Union[int, Sequence[int]]: %s" % pst)

        # UNKNOWN ARGS LEFTOVER?  Could be it got more than one of the logging args.
        if kwargs:
            raise ValueError("BaseAuth was passed unknown or redundant argument(s): %s" % kwargs)

    def setup(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
        raise NotImplementedError("setup method not implemented by %s" % self.__class__)

    def unpropagated(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
        raise NotImplementedError("unpropagated method not implemented by %s" % self.__class__)

    def clear(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
        raise NotImplementedError("clear method not implemented by %s" % self.__class__)


class HTTPProviderBase(ProviderBase):
    """
    Base class for new-model HTTP drivers

    Currently this is a null adapter, holding a place in line for any future shared
    implementation.  It may never become non-null aside from the default chal_types
    it provides, but it's a small price to pay to avoid having to stuff it in later.
    """

    def __init__(self, **kwargs: Any) -> None:
        if "chal_types" not in kwargs:
            kwargs["chal_types"] = ["http-01"]
        super().__init__(**kwargs)


class DNSProviderBase(ProviderBase):
    """
    Base class for new-model DNS drivers - legacy drivers use the one in common.py

    This is a placeholder for now, implementing only the initialization where it
    captures the upcoming alias parameter.  Implementation details for alias are TBD.
    """

    def __init__(self, **kwargs: Any) -> None:
        alias = kwargs.pop("alias", None)
        if "chal_types" not in kwargs:
            kwargs["chal_types"] = ["dns-01"]
        super().__init__(**kwargs)
        self.alias = alias

    ### minimal support for using a DNS alias to hold the challenge response

    def cname_domain(self, chal: Dict[str, str]) -> Union[str, None]:
        return "_acme-challenge." + chal["domain"] if self.alias else None

    def target_domain(self, chal: Dict[str, str]) -> str:
        d = chal["domain"]
        return "_acme-challenge." + d if not self.alias else d + "." + self.alias



"demo.py - examples of implementing non-or-minimally-functional challenge providers"

###
### NB: very early draft, minimal testing so far (May 25th)
###

from typing import Any, Dict, Sequence

from ..auth import ProviderBase
from ..dns_providers.common import dns_challenge


class ManualProvider(ProviderBase):
    def __init__(self, **kwargs: Any) -> None:
        chal_type = kwargs.pop("chal_type", "http-01")
        if chal_type not in ["dns-01", "http-01"]:
            raise ValueError("ManualProvider: invalid chal_type value: %s", chal_type)
        kwargs["chal_types"] = [chal_type]
        super().__init__(**kwargs)
        self.chal_type = chal_type

    def setup(self, challenges: Sequence[Dict[str, str]]) -> Sequence[Dict[str, str]]:
        return self._prompt("add", challenges)

    def unpropagated(self, challenges: Sequence[Dict[str, str]]) -> Sequence[Dict[str, str]]:
        # could add confirmation here, but it's just a demo
        return []

    def clear(self, challenges: Sequence[Dict[str, str]]) -> Sequence[Dict[str, str]]:
        return self._prompt("clear", challenges)

    def _prompt(self, mode: str, challenges: Sequence[Dict[str, str]]) -> Sequence[Dict[str, str]]:
        for chal in challenges:
            if self.chal_type == "dns-01":
                print(
                    "Please {0} the challenge {1} as a TXT record on _acme-validation.{ident_value}".format(
                        mode, dns_challenge(chal["key_auth"]), **chal
                    )
                )
            else:
                print(
                    "Please {0} the challenge file named {token} with contents {key_auth}".format(
                        mode, **chal
                    )
                )
        return []



import collections

try:
    route53_dependencies = True
    import boto3
    from botocore.client import Config
except ImportError:
    route53_dependencies = False

from . import common


# most code of this class is copy from certbot's route53 dns plugin.
class Route53Dns(common.BaseDns):
    ttl = 10
    connect_timeout = 30
    read_timeout = 30

    def __init__(self, access_key_id=None, secret_access_key=None):
        if not route53_dependencies:
            raise ImportError(
                """You need to install Route53Dns dependencies. run; pip3 install sewer[route53]"""
            )

        self.aws_config = Config(
            connect_timeout=self.connect_timeout, read_timeout=self.read_timeout
        )
        if access_key_id and secret_access_key:
            # use user given credential
            self.r53 = boto3.client(
                "route53",
                aws_access_key_id=access_key_id,
                aws_secret_access_key=secret_access_key,
                config=self.aws_config,
            )
        else:
            # let boto3 find credential
            # https://boto3.readthedocs.io/en/latest/guide/configuration.html#best-practices-for-configuring-credentials
            self.r53 = boto3.client("route53", config=self.aws_config)

        self._resource_records = collections.defaultdict(list)

        super(Route53Dns, self).__init__()

    def create_dns_record(self, domain_name, domain_dns_value):
        challenge_domain = "_acme-challenge" + "." + domain_name + "."
        return self._change_txt_record("UPSERT", challenge_domain, domain_dns_value)

    def delete_dns_record(self, domain_name, domain_dns_value):
        challenge_domain = "_acme-challenge" + "." + domain_name + "."
        return self._change_txt_record("DELETE", challenge_domain, domain_dns_value)

    def _find_zone_id_for_domain(self, domain):
        """Find the zone id responsible a given FQDN.
           That is, the id for the zone whose name is the longest parent of the
           domain.
        """
        paginator = self.r53.get_paginator("list_hosted_zones")
        zones = []
        target_labels = domain.rstrip(".").split(".")
        for page in paginator.paginate():
            for zone in page["HostedZones"]:
                if zone["Config"]["PrivateZone"]:
                    continue

                candidate_labels = zone["Name"].rstrip(".").split(".")
                if candidate_labels == target_labels[-len(candidate_labels) :]:
                    zones.append((zone["Name"], zone["Id"]))

        if not zones:
            raise RuntimeError("Unable to find a Route53 hosted zone for {0}".format(domain))

        # Order the zones that are suffixes for our desired to domain by
        # length, this puts them in an order like:
        # ["foo.bar.baz.com", "bar.baz.com", "baz.com", "com"]
        # And then we choose the first one, which will be the most specific.
        zones.sort(key=lambda z: len(z[0]), reverse=True)
        return zones[0][1]

    def _change_txt_record(self, action, domain_name, domain_dns_value):
        zone_id = self._find_zone_id_for_domain(domain_name)

        rrecords = self._resource_records[domain_name]
        challenge = {"Value": '"{0}"'.format(domain_dns_value)}
        if action == "DELETE":
            # Remove the record being deleted from the list of tracked records
            rrecords.remove(challenge)
            if rrecords:
                # Need to update instead, as we're not deleting the rrset
                action = "UPSERT"
            else:
                # Create a new list containing the record to use with DELETE
                rrecords = [challenge]
        else:
            rrecords.append(challenge)

        response = self.r53.change_resource_record_sets(
            HostedZoneId=zone_id,
            ChangeBatch={
                "Comment": "certbot-dns-route53 certificate validation " + action,
                "Changes": [
                    {
                        "Action": action,
                        "ResourceRecordSet": {
                            "Name": domain_name,
                            "Type": "TXT",
                            "TTL": self.ttl,
                            "ResourceRecords": rrecords,
                        },
                    }
                ],
            },
        )
        return response["ChangeInfo"]["Id"]



import urllib.parse

try:
    acmedns_dependencies = True
    from dns.resolver import Resolver
except ImportError:
    acmedns_dependencies = False
import requests

from . import common
from ..lib import log_response


class AcmeDnsDns(common.BaseDns):
    """
    """

    dns_provider_name = "acmedns"

    def __init__(self, ACME_DNS_API_USER, ACME_DNS_API_KEY, ACME_DNS_API_BASE_URL):

        if not acmedns_dependencies:
            raise ImportError(
                """You need to install AcmeDnsDns dependencies. run; pip3 install sewer[acmedns]"""
            )

        self.ACME_DNS_API_USER = ACME_DNS_API_USER
        self.ACME_DNS_API_KEY = ACME_DNS_API_KEY
        self.HTTP_TIMEOUT = 65  # seconds

        if ACME_DNS_API_BASE_URL[-1] != "/":
            self.ACME_DNS_API_BASE_URL = ACME_DNS_API_BASE_URL + "/"
        else:
            self.ACME_DNS_API_BASE_URL = ACME_DNS_API_BASE_URL
        super(AcmeDnsDns, self).__init__()

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record")

        resolver = Resolver(configure=False)
        resolver.nameservers = ["8.8.8.8"]
        answer = resolver.query("_acme-challenge.{0}.".format(domain_name), "TXT")
        subdomain, _ = str(answer.canonical_name).split(".", 1)

        url = urllib.parse.urljoin(self.ACME_DNS_API_BASE_URL, "update")
        headers = {"X-Api-User": self.ACME_DNS_API_USER, "X-Api-Key": self.ACME_DNS_API_KEY}
        body = {"subdomain": subdomain, "txt": domain_dns_value}
        update_acmedns_dns_record_response = requests.post(
            url, headers=headers, json=body, timeout=self.HTTP_TIMEOUT
        )
        self.logger.debug(
            "update_acmedns_dns_record_response. status_code={0}. response={1}".format(
                update_acmedns_dns_record_response.status_code,
                log_response(update_acmedns_dns_record_response),
            )
        )
        if update_acmedns_dns_record_response.status_code != 200:
            # raise error so that we do not continue to make calls to ACME
            # server
            raise ValueError(
                "Error creating acme-dns dns record: status_code={status_code} response={response}".format(
                    status_code=update_acmedns_dns_record_response.status_code,
                    response=log_response(update_acmedns_dns_record_response),
                )
            )
        self.logger.info("create_dns_record_end")

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record")
        # acme-dns doesn't support this
        self.logger.info("delete_dns_record_success")



import os
import argparse

from .client import Client
from . import __version__ as sewer_version
from .config import ACME_DIRECTORY_URL_STAGING, ACME_DIRECTORY_URL_PRODUCTION
from .lib import create_logger


def main():
    """
    Usage:
        1. To get a new certificate:
        CLOUDFLARE_EMAIL=example@example.com \
        CLOUDFLARE_API_KEY=api-key \
        sewer \
        --dns cloudflare \
        --domain example.com \
        --action run

        2. To renew a certificate:
        CLOUDFLARE_EMAIL=example@example.com \
        CLOUDFLARE_API_KEY=api-key \
        sewer \
        --account_key /path/to/your/account.key \
        --dns cloudflare \
        --domain example.com \
        --action renew
    """
    parser = argparse.ArgumentParser(
        prog="sewer",
        description="""Sewer is a Let's Encrypt(ACME) client.
            Example usage::
            CLOUDFLARE_EMAIL=example@example.com \
            CLOUDFLARE_API_KEY=api-key \
            sewer \
            --dns cloudflare \
            --domain example.com \
            --action run""",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=sewer_version.__version__),
        help="The currently installed sewer version.",
    )
    parser.add_argument(
        "--account_key",
        type=argparse.FileType("r"),
        required=False,
        help="The path to your letsencrypt/acme account key. \
        eg: --account_key /home/myaccount.key",
    )
    parser.add_argument(
        "--certificate_key",
        type=argparse.FileType("r"),
        required=False,
        help="The path to your certificate key. \
        eg: --certificate_key /home/mycertificate.key",
    )
    parser.add_argument(
        "--dns",
        type=str,
        required=True,
        choices=[
            "cloudflare",
            "aurora",
            "acmedns",
            "aliyun",
            "hurricane",
            "rackspace",
            "dnspod",
            "duckdns",
            "cloudns",
            "powerdns",
            "gandi",
        ],
        help="The name of the dns provider that you want to use.",
    )
    parser.add_argument(
        "--domain",
        type=str,
        required=True,
        help="The domain/subdomain name for which \
        you want to get/renew certificate for. \
        wildcards are also supported \
        eg: --domain example.com",
    )
    parser.add_argument(
        "--alt_domains",
        type=str,
        required=False,
        default=[],
        nargs="*",
        help="A list of alternative domain/subdomain name/s(if any) for which \
        you want to get/renew certificate for. \
        eg: --alt_domains www.example.com blog.example.com",
    )
    parser.add_argument(
        "--bundle_name",
        type=str,
        required=False,
        help="The name to use for certificate \
        certificate key and account key. Default is name of domain.",
    )
    parser.add_argument(
        "--endpoint",
        type=str,
        required=False,
        default="production",
        choices=["production", "staging"],
        help="Whether to use letsencrypt/acme production/live endpoints \
        or staging endpoints. production endpoints are used by default. \
        eg: --endpoint staging",
    )
    parser.add_argument(
        "--email",
        type=str,
        required=False,
        help="Email to be used for registration and recovery. \
        eg: --email me@example.com",
    )
    parser.add_argument(
        "--action",
        type=str,
        required=True,
        choices=["run", "renew"],
        help="The action that you want to perform. \
        Either run (get a new certificate) or renew (renew a certificate). \
        eg: --action run",
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        required=False,
        default=os.getcwd(),
        help="""The dir where the certificate and keys file will be stored.
            default:  The directory you run sewer command.
            eg: --out_dir /data/ssl/
            """,
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        required=False,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="The log level to output log messages at. \
        eg: --loglevel DEBUG",
    )

    args = parser.parse_args()

    dns_provider = args.dns
    domain = args.domain
    alt_domains = args.alt_domains
    action = args.action
    account_key = args.account_key
    certificate_key = args.certificate_key
    bundle_name = args.bundle_name
    endpoint = args.endpoint
    email = args.email
    loglevel = args.loglevel
    out_dir = args.out_dir

    # Make sure the output dir user specified is writable
    if not os.access(out_dir, os.W_OK):
        raise OSError("The dir '{0}' is not writable".format(out_dir))

    logger = create_logger(None, loglevel)

    if account_key:
        account_key = account_key.read()
    if certificate_key:
        certificate_key = certificate_key.read()
    if bundle_name:
        file_name = bundle_name
    else:
        file_name = "{0}".format(domain)
    if endpoint == "staging":
        ACME_DIRECTORY_URL = ACME_DIRECTORY_URL_STAGING
    else:
        ACME_DIRECTORY_URL = ACME_DIRECTORY_URL_PRODUCTION

    if dns_provider == "cloudflare":
        from . import CloudFlareDns

        CLOUDFLARE_EMAIL = os.environ.get("CLOUDFLARE_EMAIL", None)
        CLOUDFLARE_API_KEY = os.environ.get("CLOUDFLARE_API_KEY", None)
        CLOUDFLARE_TOKEN = os.environ.get("CLOUDFLARE_TOKEN", None)

        if CLOUDFLARE_EMAIL and CLOUDFLARE_API_KEY and not CLOUDFLARE_TOKEN:
            dns_class = CloudFlareDns(
                CLOUDFLARE_EMAIL=CLOUDFLARE_EMAIL, CLOUDFLARE_API_KEY=CLOUDFLARE_API_KEY
            )
        elif CLOUDFLARE_TOKEN and not CLOUDFLARE_EMAIL and not CLOUDFLARE_API_KEY:
            dns_class = CloudFlareDns(CLOUDFLARE_TOKEN=CLOUDFLARE_TOKEN)
        else:
            err = (
                "ERROR:: Please supply either CLOUDFLARE_EMAIL and CLOUDFLARE_API_KEY"
                "or CLOUDFLARE_TOKEN as environment variables."
            )
            logger.error(err)
            raise KeyError(err)

        logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))

    elif dns_provider == "aurora":
        from . import AuroraDns

        try:
            AURORA_API_KEY = os.environ["AURORA_API_KEY"]
            AURORA_SECRET_KEY = os.environ["AURORA_SECRET_KEY"]

            dns_class = AuroraDns(
                AURORA_API_KEY=AURORA_API_KEY, AURORA_SECRET_KEY=AURORA_SECRET_KEY
            )
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise

    elif dns_provider == "acmedns":
        from . import AcmeDnsDns

        try:
            ACME_DNS_API_USER = os.environ["ACME_DNS_API_USER"]
            ACME_DNS_API_KEY = os.environ["ACME_DNS_API_KEY"]
            ACME_DNS_API_BASE_URL = os.environ["ACME_DNS_API_BASE_URL"]

            dns_class = AcmeDnsDns(
                ACME_DNS_API_USER=ACME_DNS_API_USER,
                ACME_DNS_API_KEY=ACME_DNS_API_KEY,
                ACME_DNS_API_BASE_URL=ACME_DNS_API_BASE_URL,
            )
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "aliyun":
        from . import AliyunDns

        try:
            aliyun_ak = os.environ["ALIYUN_AK_ID"]
            aliyun_secret = os.environ["ALIYUN_AK_SECRET"]
            aliyun_endpoint = os.environ.get("ALIYUN_ENDPOINT", "cn-beijing")
            dns_class = AliyunDns(aliyun_ak, aliyun_secret, aliyun_endpoint)
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "hurricane":
        from . import HurricaneDns

        try:
            he_username = os.environ["HURRICANE_USERNAME"]
            he_password = os.environ["HURRICANE_PASSWORD"]
            dns_class = HurricaneDns(he_username, he_password)
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "rackspace":
        from . import RackspaceDns

        try:
            RACKSPACE_USERNAME = os.environ["RACKSPACE_USERNAME"]
            RACKSPACE_API_KEY = os.environ["RACKSPACE_API_KEY"]
            dns_class = RackspaceDns(RACKSPACE_USERNAME, RACKSPACE_API_KEY)
            logger.info("chosen_dns_prover. Using {0} as dns provider. ".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "dnspod":
        from . import DNSPodDns

        try:
            DNSPOD_ID = os.environ["DNSPOD_ID"]
            DNSPOD_API_KEY = os.environ["DNSPOD_API_KEY"]
            dns_class = DNSPodDns(DNSPOD_ID, DNSPOD_API_KEY)
            logger.info("chosen_dns_prover. Using {0} as dns provider. ".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "duckdns":
        from . import DuckDNSDns

        try:
            duckdns_token = os.environ["DUCKDNS_TOKEN"]

            dns_class = DuckDNSDns(duckdns_token=duckdns_token)
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "cloudns":
        from . import ClouDNSDns

        try:
            dns_class = ClouDNSDns()
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "powerdns":
        from . import PowerDNSDns

        try:
            powerdns_api_key = os.environ["POWERDNS_API_KEY"]
            powerdns_api_url = os.environ["POWERDNS_API_URL"]

            dns_class = PowerDNSDns(powerdns_api_key, powerdns_api_url)
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    elif dns_provider == "gandi":
        from . import GandiDns

        try:
            gandi_api_key = os.environ["GANDI_API_KEY"]

            dns_class = GandiDns(GANDI_API_KEY=gandi_api_key)
            logger.info("chosen_dns_provider. Using {0} as dns provider.".format(dns_provider))
        except KeyError as e:
            logger.error("ERROR:: Please supply {0} as an environment variable.".format(str(e)))
            raise
    else:
        raise ValueError("The dns provider {0} is not recognised.".format(dns_provider))

    client = Client(
        domain_name=domain,
        dns_class=dns_class,
        domain_alt_names=alt_domains,
        contact_email=email,
        account_key=account_key,
        certificate_key=certificate_key,
        ACME_DIRECTORY_URL=ACME_DIRECTORY_URL,
        LOG_LEVEL=loglevel,
    )
    certificate_key = client.certificate_key
    account_key = client.account_key

    # prepare file path
    account_key_file_path = os.path.join(out_dir, "{0}.account.key".format(file_name))
    crt_file_path = os.path.join(out_dir, "{0}.crt".format(file_name))
    crt_key_file_path = os.path.join(out_dir, "{0}.key".format(file_name))

    # write out account_key in out_dir directory
    with open(account_key_file_path, "w") as account_file:
        account_file.write(account_key)
    logger.info("account key succesfully written to {0}.".format(account_key_file_path))

    if action == "renew":
        message = "Certificate Succesfully renewed. The certificate, certificate key and account key have been saved in the current directory"
        certificate = client.renew()
    else:
        message = "Certificate Succesfully issued. The certificate, certificate key and account key have been saved in the current directory"
        certificate = client.cert()

    # write out certificate and certificate key in out_dir directory
    with open(crt_file_path, "w") as certificate_file:
        certificate_file.write(certificate)
    with open(crt_key_file_path, "w") as certificate_key_file:
        certificate_key_file.write(certificate_key)

    logger.info("certificate succesfully written to {0}.".format(crt_file_path))
    logger.info("certificate key succesfully written to {0}.".format(crt_key_file_path))

    logger.info("the_end. {0}".format(message))



from hashlib import sha256
from typing import Any, Dict, Sequence

from sewer.auth import ErrataItemType, ProviderBase
from sewer.lib import safe_base64


def dns_challenge(key_auth: str) -> str:
    "return safe-base64 of hash of key_auth; used for dns response"

    return safe_base64(sha256(key_auth.encode("utf8")).digest())


class BaseDns(ProviderBase):
    """
    Shim for legacy DNS provider interface.
    """

    def __init__(self, **kwargs: Any) -> None:
        if "chal_types" not in kwargs:
            kwargs["chal_types"] = ["dns-01"]
        super().__init__(**kwargs)

    ### shim methods

    def setup(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
        for chal in challenges:
            self.create_dns_record(chal["ident_value"], dns_challenge(chal["key_auth"]))
        return []

    def unpropagated(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
        return []

    def clear(self, challenges: Sequence[Dict[str, str]]) -> Sequence[ErrataItemType]:
        for chal in challenges:
            self.delete_dns_record(chal["ident_value"], dns_challenge(chal["key_auth"]))
        return []

    ### legacy DNS methods

    def create_dns_record(self, domain_name, domain_dns_value):
        """
        Method that creates/adds a dns TXT record for a domain/subdomain name on
        a chosen DNS provider.

        :param domain_name: :string: The domain/subdomain name whose dns record ought to be
            created/added on a chosen DNS provider.
        :param domain_dns_value: :string: The value/content of the TXT record that will be
            created/added for the given domain/subdomain

        This method should return None

        Basic Usage:
            If the value of the `domain_name` variable is example.com and the value of
            `domain_dns_value` is HAJA_4MkowIFByHhFaP8u035skaM91lTKplKld
            Then, your implementation of this method ought to create a DNS TXT record
            whose name is '_acme-challenge' + '.' + domain_name + '.' (ie: _acme-challenge.example.com. )
            and whose value/content is HAJA_4MkowIFByHhFaP8u035skaM91lTKplKld

            Using a dns client like dig(https://linux.die.net/man/1/dig) to do a dns lookup should result
            in something like:
                dig TXT _acme-challenge.example.com
                ...
                ;; ANSWER SECTION:
                _acme-challenge.example.com. 120 IN TXT "HAJA_4MkowIFByHhFaP8u035skaM91lTKplKld"
                _acme-challenge.singularity.brandur.org. 120 IN TXT "9C0DqKC_4MkowIFByHhFaP8u0Zv4z7Wz2IHM91lTKec"
            Optionally, you may also use an online dns client like: https://toolbox.googleapps.com/apps/dig/#TXT/

            Please consult your dns provider on how/format of their DNS TXT records.
            You may also want to consult the cloudflare DNS implementation that is found in this repository.
        """
        raise NotImplementedError("create_dns_record method must be implemented.")

    def delete_dns_record(self, domain_name, domain_dns_value):
        """
        Method that deletes/removes a dns TXT record for a domain/subdomain name on
        a chosen DNS provider.

        :param domain_name: :string: The domain/subdomain name whose dns record ought to be
            deleted/removed on a chosen DNS provider.
        :param domain_dns_value: :string: The value/content of the TXT record that will be
            deleted/removed for the given domain/subdomain

        This method should return None
        """
        raise NotImplementedError("delete_dns_record method must be implemented.")



import urllib.parse

import requests

from . import common


class DNSPodDns(common.BaseDns):
    """
    """

    dns_provider_name = "dnspod"

    def __init__(self, DNSPOD_ID, DNSPOD_API_KEY, DNSPOD_API_BASE_URL="https://dnsapi.cn/"):
        self.DNSPOD_ID = DNSPOD_ID
        self.DNSPOD_API_KEY = DNSPOD_API_KEY
        self.DNSPOD_API_BASE_URL = DNSPOD_API_BASE_URL
        self.HTTP_TIMEOUT = 65  # seconds
        self.DNSPOD_LOGIN = "{0},{1}".format(self.DNSPOD_ID, self.DNSPOD_API_KEY)

        if DNSPOD_API_BASE_URL[-1] != "/":
            self.DNSPOD_API_BASE_URL = DNSPOD_API_BASE_URL + "/"
        else:
            self.DNSPOD_API_BASE_URL = DNSPOD_API_BASE_URL
        super(DNSPodDns, self).__init__()

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record")
        subd = ""

        ### FIX ME ### domain is exactly last two parts (no bbc.co.uk eg.)

        if domain_name.count(".") != 1:  # not top level domain
            pos = domain_name.rfind(".", 0, domain_name.rfind("."))
            subd = domain_name[:pos]
            domain_name = domain_name[pos + 1 :]
            if subd != "":
                subd = "." + subd

        url = urllib.parse.urljoin(self.DNSPOD_API_BASE_URL, "Record.Create")
        body = {
            "record_type": "TXT",
            "domain": domain_name,
            "sub_domain": "_acme-challenge" + subd,
            "value": domain_dns_value,
            "record_line_id": "0",
            "format": "json",
            "login_token": self.DNSPOD_LOGIN,
        }
        create_dnspod_dns_record_response = requests.post(
            url, data=body, timeout=self.HTTP_TIMEOUT
        ).json()
        self.logger.debug(
            "create_dnspod_dns_record_response. status_code={0}. response={1}".format(
                create_dnspod_dns_record_response["status"]["code"],
                create_dnspod_dns_record_response["status"]["message"],
            )
        )
        if create_dnspod_dns_record_response["status"]["code"] != "1":
            # raise error so that we do not continue to make calls to ACME
            # server
            raise ValueError(
                "Error creating dnspod dns record: status_code={status_code} response={response}".format(
                    status_code=create_dnspod_dns_record_response["status"]["code"],
                    response=create_dnspod_dns_record_response["status"]["message"],
                )
            )
        self.logger.info("create_dns_record_end")

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record")
        subd = ""

        ### FIX ME ### domain is exactly last two parts (no bbc.co.uk eg.)

        if domain_name.count(".") != 1:  # not top level domain
            pos = domain_name.rfind(".", 0, domain_name.rfind("."))
            subd = domain_name[:pos]
            domain_name = domain_name[pos + 1 :]
            if subd != "":
                subd = "." + subd

        url = urllib.parse.urljoin(self.DNSPOD_API_BASE_URL, "Record.List")
        # pos = domain_name.rfind(".",0, domain_name.rfind("."))
        subdomain = "_acme-challenge." + subd
        rootdomain = domain_name
        body = {
            "login_token": self.DNSPOD_LOGIN,
            "format": "json",
            "domain": rootdomain,
            "subdomain": subdomain,
            "record_type": "TXT",
        }
        list_dns_response = requests.post(url, data=body, timeout=self.HTTP_TIMEOUT).json()
        if list_dns_response["status"]["code"] != "1":
            self.logger.error(
                "list_dns_record_response. status_code={0}. message={1}".format(
                    list_dns_response["status"]["code"], list_dns_response["status"]["message"]
                )
            )
        for i in range(0, len(list_dns_response["records"])):
            rid = list_dns_response["records"][i]["id"]
            urlr = urllib.parse.urljoin(self.DNSPOD_API_BASE_URL, "Record.Remove")
            bodyr = {
                "login_token": self.DNSPOD_LOGIN,
                "format": "json",
                "domain": rootdomain,
                "record_id": rid,
            }
            delete_dns_record_response = requests.post(
                urlr, data=bodyr, timeout=self.HTTP_TIMEOUT
            ).json()
            if delete_dns_record_response["status"]["code"] != "1":
                self.logger.error(
                    "delete_dns_record_response. status_code={0}. message={1}".format(
                        delete_dns_record_response["status"]["code"],
                        delete_dns_record_response["status"]["message"],
                    )
                )

        self.logger.info("delete_dns_record_success")



"""
Hurricane Electric DNS Support
"""
import json

try:
    hedns_dependencies = True
    import HurricaneDNS as _hurricanedns
except ImportError:
    hedns_dependencies = False

from . import common


class _Response(object):
    """
    wrapper aliyun resp to the format sewer wanted.
    """

    def __init__(self, status_code=200, content=None, headers=None):
        self.status_code = status_code
        self.headers = headers or {}
        self.content = content or {}
        self.content = json.dumps(content)
        super(_Response, self).__init__()

    def json(self):
        return json.loads(self.content)


class HurricaneDns(common.BaseDns):
    def __init__(self, username, password):
        super(HurricaneDns, self).__init__()
        if not hedns_dependencies:
            raise ImportError(
                """You need to install HurricaneDns dependencies. run: pip3 install sewer[hurricane]"""
            )

        self.clt = _hurricanedns.HurricaneDNS(username, password)

    @staticmethod
    def extract_zone(domain_name):
        """
        extract domain to root, sub, acme_txt
        :param str domain_name: the value sewer client passed in, like *.menduo.example.com
        :return tuple: root, zone, acme_txt
        """
        if domain_name.count(".") > 1:
            zone, middle, last = str(domain_name).rsplit(".", 2)
            root = ".".join([middle, last])
            acme_txt = "_acme-challenge.%s" % zone
        else:
            zone = ""
            root = domain_name
            acme_txt = "_acme-challenge"
        return root, zone, acme_txt

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record start: %s", (domain_name, domain_dns_value))

        root, _, acme_txt = self.extract_zone(domain_name)
        self.clt.add_record(root, acme_txt, "TXT", domain_dns_value, ttl=300)

        self.logger.info("create_dns_record end: %s", (domain_name, domain_dns_value))

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record start: %s", (domain_name, domain_dns_value))

        root, _, acme_txt = self.extract_zone(domain_name)
        host = "%s.%s" % (acme_txt, root)

        recored_list = self.clt.get_records(root, host, "TXT")

        for i in recored_list:
            self.clt.del_record(root, i["id"])

        self.logger.info("delete_dns_record end: %s", (domain_name, domain_dns_value))



# DNS Provider for AuroRa DNS from the dutch hosting provider pcextreme
# https://www.pcextreme.nl/aurora/dns
# Aurora uses libcloud from apache
# https://libcloud.apache.org/
try:
    aurora_dependencies = True
    from libcloud.dns.providers import get_driver
    from libcloud.dns.types import Provider, RecordType
    import tldextract
except ImportError:
    aurora_dependencies = False
from . import common


class AuroraDns(common.BaseDns):
    """
    Todo: re-organize this class so that we make it easier to mock things out to
    facilitate better tests.
    """

    dns_provider_name = "aurora"

    def __init__(self, AURORA_API_KEY, AURORA_SECRET_KEY):

        if not aurora_dependencies:
            raise ImportError(
                """You need to install AuroraDns dependencies. run; pip3 install sewer[aurora]"""
            )

        self.AURORA_API_KEY = AURORA_API_KEY
        self.AURORA_SECRET_KEY = AURORA_SECRET_KEY
        super(AuroraDns, self).__init__()

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record")

        extractedDomain = tldextract.extract(domain_name)
        domainSuffix = extractedDomain.domain + "." + extractedDomain.suffix

        if extractedDomain.subdomain is "":
            subDomain = "_acme-challenge"
        else:
            subDomain = "_acme-challenge." + extractedDomain.subdomain

        cls = get_driver(Provider.AURORADNS)
        driver = cls(key=self.AURORA_API_KEY, secret=self.AURORA_SECRET_KEY)
        zone = driver.get_zone(domainSuffix)
        zone.create_record(name=subDomain, type=RecordType.TXT, data=domain_dns_value)

        self.logger.info("create_dns_record_success")
        return

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record")

        extractedDomain = tldextract.extract(domain_name)
        domainSuffix = extractedDomain.domain + "." + extractedDomain.suffix

        if extractedDomain.subdomain is "":
            subDomain = "_acme-challenge"
        else:
            subDomain = "_acme-challenge." + extractedDomain.subdomain

        cls = get_driver(Provider.AURORADNS)
        driver = cls(key=self.AURORA_API_KEY, secret=self.AURORA_SECRET_KEY)
        zone = driver.get_zone(domainSuffix)

        records = driver.list_records(zone)
        for x in records:
            if x.name == subDomain and x.type == "TXT":
                record_id = x.id
                self.logger.info(
                    "Found record "
                    + subDomain
                    + "."
                    + domainSuffix
                    + " with id : "
                    + record_id
                    + "."
                )
                record = driver.get_record(zone_id=zone.id, record_id=record_id)
                driver.delete_record(record)
                self.logger.info(
                    "Deleted record "
                    + subDomain
                    + "."
                    + domainSuffix
                    + " with id : "
                    + record_id
                    + "."
                )
            else:
                self.logger.info(
                    "Record " + subDomain + "." + domainSuffix + " not found. No record to delete."
                )

        self.logger.info("delete_dns_record_success")
        return



import urllib.parse
import requests
from . import common
from ..lib import log_response

try:
    rackspace_dependencies = True
    import tldextract
except ImportError:
    rackspace_dependencies = False

import time


class RackspaceDns(common.BaseDns):
    """
    """

    dns_providername = "rackspace"

    def get_rackspace_credentials(self):
        self.logger.debug("get_rackspace_credentials")
        RACKSPACE_IDENTITY_URL = "https://identity.api.rackspacecloud.com/v2.0/tokens"
        payload = {
            "auth": {
                "RAX-KSKEY:apiKeyCredentials": {
                    "username": self.RACKSPACE_USERNAME,
                    "apiKey": self.RACKSPACE_API_KEY,
                }
            }
        }
        find_rackspace_api_details_response = requests.post(RACKSPACE_IDENTITY_URL, json=payload)
        self.logger.debug(
            "find_rackspace_api_details_response. status_code={0}".format(
                find_rackspace_api_details_response.status_code
            )
        )
        if find_rackspace_api_details_response.status_code != 200:
            raise ValueError(
                "Error getting token and URL details from rackspace identity server: status_code={status_code} response={response}".format(
                    status_code=find_rackspace_api_details_response.status_code,
                    response=log_response(find_rackspace_api_details_response),
                )
            )
        data = find_rackspace_api_details_response.json()
        api_token = data["access"]["token"]["id"]
        url_data = next(
            (item for item in data["access"]["serviceCatalog"] if item["type"] == "rax:dns"), None
        )
        if url_data is None:
            raise ValueError(
                "Error finding url data for the rackspace dns api in the response from the identity server"
            )
        else:
            api_base_url = url_data["endpoints"][0]["publicURL"] + "/"
        return (api_token, api_base_url)

    def __init__(self, RACKSPACE_USERNAME, RACKSPACE_API_KEY):

        if not rackspace_dependencies:
            raise ImportError(
                """You need to install RackspaceDns dependencies. run; pip3 install sewer[rackspace]"""
            )
        self.RACKSPACE_DNS_ZONE_ID = None
        self.RACKSPACE_USERNAME = RACKSPACE_USERNAME
        self.RACKSPACE_API_KEY = RACKSPACE_API_KEY
        self.HTTP_TIMEOUT = 65  # seconds
        super(RackspaceDns, self).__init__()
        self.RACKSPACE_API_TOKEN, self.RACKSPACE_API_BASE_URL = self.get_rackspace_credentials()
        self.RACKSPACE_HEADERS = {
            "X-Auth-Token": self.RACKSPACE_API_TOKEN,
            "Content-Type": "application/json",
        }

    def get_dns_zone(self, domain_name):
        self.logger.debug("get_dns_zone")
        extracted_domain = tldextract.extract(domain_name)
        self.RACKSPACE_DNS_ZONE = ".".join([extracted_domain.domain, extracted_domain.suffix])

    def find_dns_zone_id(self, domain_name):
        self.logger.debug("find_dns_zone_id")
        self.get_dns_zone(domain_name)
        url = self.RACKSPACE_API_BASE_URL + "domains"
        find_dns_zone_id_response = requests.get(url, headers=self.RACKSPACE_HEADERS)
        self.logger.debug(
            "find_dns_zone_id_response. status_code={0}".format(
                find_dns_zone_id_response.status_code
            )
        )
        if find_dns_zone_id_response.status_code != 200:
            raise ValueError(
                "Error getting rackspace dns domain info: status_code={status_code} response={response}".format(
                    status_code=find_dns_zone_id_response.status_code,
                    response=log_response(find_dns_zone_id_response),
                )
            )
        result = find_dns_zone_id_response.json()
        domain_data = next(
            (item for item in result["domains"] if item["name"] == self.RACKSPACE_DNS_ZONE), None
        )
        if domain_data is None:
            raise ValueError(
                "Error finding information for {dns_zone} in dns response data:\n{response_data})".format(
                    dns_zone=self.RACKSPACE_DNS_ZONE,
                    response_data=log_response(find_dns_zone_id_response),
                )
            )
        dns_zone_id = domain_data["id"]
        self.logger.debug("find_dns_zone_id_success")
        return dns_zone_id

    def find_dns_record_id(self, domain_name, domain_dns_value):
        self.logger.debug("find_dns_record_id")
        self.RACKSPACE_DNS_ZONE_ID = self.find_dns_zone_id(domain_name)
        url = self.RACKSPACE_API_BASE_URL + "domains/{0}/records".format(self.RACKSPACE_DNS_ZONE_ID)
        find_dns_record_id_response = requests.get(url, headers=self.RACKSPACE_HEADERS)
        self.logger.debug(
            "find_dns_record_id_response. status_code={0}".format(
                find_dns_record_id_response.status_code
            )
        )
        self.logger.debug(url)
        if find_dns_record_id_response.status_code != 200:
            raise ValueError(
                "Error finding dns records for {dns_zone}: status_code={status_code} response={response}".format(
                    dns_zone=self.RACKSPACE_DNS_ZONE,
                    status_code=find_dns_record_id_response.status_code,
                    response=log_response(find_dns_record_id_response),
                )
            )
        records = find_dns_record_id_response.json()["records"]
        RACKSPACE_RECORD_DATA = next(
            (item for item in records if item["data"] == domain_dns_value), None
        )
        if RACKSPACE_RECORD_DATA is None:
            raise ValueError(
                "Couldn't find record with name {domain_name}\ncontaining data: {domain_dns_value}\nin the response data:{response_data}".format(
                    domain_name=domain_name,
                    domain_dns_value=domain_dns_value,
                    response_data=log_response(find_dns_record_id_response),
                )
            )
        record_id = RACKSPACE_RECORD_DATA["id"]
        self.logger.debug("find_dns_record_id success")
        return record_id

    def poll_callback_url(self, callback_url):
        start_time = time.time()
        while True:
            callback_url_response = requests.get(callback_url, headers=self.RACKSPACE_HEADERS)
            if time.time() > start_time + self.HTTP_TIMEOUT:
                raise ValueError(
                    "Timed out polling callbackurl for dns record status.  Last status_code={status_code} last response={response}".format(
                        status_code=callback_url_response.status_code,
                        response=log_response(callback_url_response),
                    )
                )
            if callback_url_response.status_code != 200:
                raise Exception(
                    "Could not get dns record status from callback url.  Status code ={status_code}. response={response}".format(
                        status_code=callback_url_response.status_code,
                        response=log_response(callback_url_response),
                    )
                )
            if callback_url_response.json()["status"] == "ERROR":
                raise Exception(
                    "Error in creating/deleting dns record: status_Code={status_code}. response={response}".format(
                        status_code=callback_url_response.status_code,
                        response=log_response(callback_url_response),
                    )
                )
            if callback_url_response.json()["status"] == "COMPLETED":
                break

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record")
        self.RACKSPACE_DNS_ZONE_ID = self.find_dns_zone_id(domain_name)
        record_name = "_acme-challenge." + domain_name
        url = urllib.parse.urljoin(
            self.RACKSPACE_API_BASE_URL, "domains/{0}/records".format(self.RACKSPACE_DNS_ZONE_ID)
        )
        body = {
            "records": [{"name": record_name, "type": "TXT", "data": domain_dns_value, "ttl": 3600}]
        }
        create_rackspace_dns_record_response = requests.post(
            url, headers=self.RACKSPACE_HEADERS, json=body, timeout=self.HTTP_TIMEOUT
        )
        self.logger.debug(
            "create_rackspace_dns_record_response. status_code={status_code}".format(
                status_code=create_rackspace_dns_record_response.status_code
            )
        )
        if create_rackspace_dns_record_response.status_code != 202:
            raise ValueError(
                "Error creating rackspace dns record: status_code={status_code} response={response}".format(
                    status_code=create_rackspace_dns_record_response.status_code,
                    response=create_rackspace_dns_record_response.text,
                )
            )
            # response=log_response(create_rackspace_dns_record_response)))
            # After posting the dns record we want created, the response gives us a url to check that will
        # update when the job is done
        callback_url = create_rackspace_dns_record_response.json()["callbackUrl"]
        self.poll_callback_url(callback_url)
        self.logger.info(
            "create_dns_record_success. Name: {record_name} Data: {data}".format(
                record_name=record_name, data=domain_dns_value
            )
        )

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record")
        record_name = "_acme-challenge." + domain_name
        self.RACKSPACE_DNS_ZONE_ID = self.find_dns_zone_id(domain_name)
        self.RACKSPACE_RECORD_ID = self.find_dns_record_id(domain_name, domain_dns_value)
        url = self.RACKSPACE_API_BASE_URL + "domains/{domain_id}/records/?id={record_id}".format(
            domain_id=self.RACKSPACE_DNS_ZONE_ID, record_id=self.RACKSPACE_RECORD_ID
        )
        delete_dns_record_response = requests.delete(url, headers=self.RACKSPACE_HEADERS)
        # After sending a delete request, if all goes well, we get a 202 from the server and a URL that we can poll
        # to see when the job is done
        self.logger.debug(
            "delete_dns_record_response={0}".format(delete_dns_record_response.status_code)
        )
        if delete_dns_record_response.status_code != 202:
            raise ValueError(
                "Error deleting rackspace dns record: status_code={status_code} response={response}".format(
                    status_code=delete_dns_record_response.status_code,
                    response=log_response(delete_dns_record_response),
                )
            )
        callback_url = delete_dns_record_response.json()["callbackUrl"]
        self.poll_callback_url(callback_url)
        self.logger.info(
            "delete_dns_record_success. Name: {record_name} Data: {data}".format(
                record_name=record_name, data=domain_dns_value
            )
        )



import urllib.parse

import requests

from . import common
from ..lib import log_response


class CloudFlareDns(common.BaseDns):
    """
    """

    dns_provider_name = "cloudflare"

    def __init__(
        self,
        CLOUDFLARE_EMAIL=None,
        CLOUDFLARE_API_KEY=None,
        CLOUDFLARE_API_BASE_URL="https://api.cloudflare.com/client/v4/",
        CLOUDFLARE_TOKEN=None,
    ):
        self.CLOUDFLARE_DNS_ZONE_ID = None
        self.CLOUDFLARE_EMAIL = CLOUDFLARE_EMAIL
        self.CLOUDFLARE_API_KEY = CLOUDFLARE_API_KEY
        self.CLOUDFLARE_API_BASE_URL = CLOUDFLARE_API_BASE_URL
        self.CLOUDFLARE_TOKEN = CLOUDFLARE_TOKEN
        self.HTTP_TIMEOUT = 65  # seconds

        if CLOUDFLARE_API_BASE_URL[-1] != "/":
            self.CLOUDFLARE_API_BASE_URL = CLOUDFLARE_API_BASE_URL + "/"
        else:
            self.CLOUDFLARE_API_BASE_URL = CLOUDFLARE_API_BASE_URL

        # Either only pass a token or only the email and API key
        if not (
            (CLOUDFLARE_TOKEN and not CLOUDFLARE_EMAIL and not CLOUDFLARE_API_KEY)
            or (not CLOUDFLARE_TOKEN and CLOUDFLARE_EMAIL and CLOUDFLARE_API_KEY)
        ):
            raise ValueError(
                "Error initializing Cloudflare DNS adapter. Pass either email and API key or a token."
            )

        super(CloudFlareDns, self).__init__()

    def find_dns_zone(self, domain_name):
        self.logger.debug("find_dns_zone")
        url = urllib.parse.urljoin(self.CLOUDFLARE_API_BASE_URL, "zones?status=active")
        headers = self._get_auth_header()
        find_dns_zone_response = requests.get(url, headers=headers, timeout=self.HTTP_TIMEOUT)
        self.logger.debug(
            "find_dns_zone_response. status_code={0}".format(find_dns_zone_response.status_code)
        )
        if find_dns_zone_response.status_code != 200:
            raise ValueError(
                "Error creating cloudflare dns record: status_code={status_code} response={response}".format(
                    status_code=find_dns_zone_response.status_code,
                    response=log_response(find_dns_zone_response),
                )
            )

        result = find_dns_zone_response.json()["result"]
        for i in result:
            if i["name"] in domain_name:
                setattr(self, "CLOUDFLARE_DNS_ZONE_ID", i["id"])
        if isinstance(self.CLOUDFLARE_DNS_ZONE_ID, type(None)):
            raise ValueError(
                "Error unable to get DNS zone for domain_name={domain_name}: status_code={status_code} response={response}".format(
                    domain_name=domain_name,
                    status_code=find_dns_zone_response.status_code,
                    response=log_response(find_dns_zone_response),
                )
            )

        self.logger.debug("find_dns_zone_success")

    def create_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("create_dns_record")
        self.find_dns_zone(domain_name)

        url = urllib.parse.urljoin(
            self.CLOUDFLARE_API_BASE_URL,
            "zones/{0}/dns_records".format(self.CLOUDFLARE_DNS_ZONE_ID),
        )
        headers = self._get_auth_header()
        body = {
            "type": "TXT",
            "name": "_acme-challenge" + "." + domain_name + ".",
            "content": "{0}".format(domain_dns_value),
        }
        create_cloudflare_dns_record_response = requests.post(
            url, headers=headers, json=body, timeout=self.HTTP_TIMEOUT
        )
        self.logger.debug(
            "create_cloudflare_dns_record_response. status_code={0}. response={1}".format(
                create_cloudflare_dns_record_response.status_code,
                log_response(create_cloudflare_dns_record_response),
            )
        )
        if create_cloudflare_dns_record_response.status_code != 200:
            # raise error so that we do not continue to make calls to ACME
            # server
            raise ValueError(
                "Error creating cloudflare dns record: status_code={status_code} response={response}".format(
                    status_code=create_cloudflare_dns_record_response.status_code,
                    response=log_response(create_cloudflare_dns_record_response),
                )
            )
        self.logger.info("create_dns_record_end")

    def delete_dns_record(self, domain_name, domain_dns_value):
        self.logger.info("delete_dns_record")

        class MockResponse(object):
            def __init__(self, status_code=200, content="mock-response"):
                self.status_code = status_code
                self.content = content
                super(MockResponse, self).__init__()

            def json(self):
                return {}

        delete_dns_record_response = MockResponse()
        headers = self._get_auth_header()

        dns_name = "_acme-challenge" + "." + domain_name
        list_dns_payload = {"type": "TXT", "name": dns_name}
        list_dns_url = urllib.parse.urljoin(
            self.CLOUDFLARE_API_BASE_URL,
            "zones/{0}/dns_records".format(self.CLOUDFLARE_DNS_ZONE_ID),
        )

        list_dns_response = requests.get(
            list_dns_url, params=list_dns_payload, headers=headers, timeout=self.HTTP_TIMEOUT
        )

        for i in range(0, len(list_dns_response.json()["result"])):
            dns_record_id = list_dns_response.json()["result"][i]["id"]
            url = urllib.parse.urljoin(
                self.CLOUDFLARE_API_BASE_URL,
                "zones/{0}/dns_records/{1}".format(self.CLOUDFLARE_DNS_ZONE_ID, dns_record_id),
            )
            headers = self._get_auth_header()
            delete_dns_record_response = requests.delete(
                url, headers=headers, timeout=self.HTTP_TIMEOUT
            )
            self.logger.debug(
                "delete_dns_record_response. status_code={0}. response={1}".format(
                    delete_dns_record_response.status_code, log_response(delete_dns_record_response)
                )
            )
            if delete_dns_record_response.status_code != 200:
                # extended logging for debugging
                # we do not need to raise exception
                self.logger.error(
                    "delete_dns_record_response. status_code={0}. response={1}".format(
                        delete_dns_record_response.status_code,
                        log_response(delete_dns_record_response),
                    )
                )

        self.logger.info("delete_dns_record_success")

    def _get_auth_header(self):
        if self.CLOUDFLARE_TOKEN:
            return {"Authorization": "Bearer " + self.CLOUDFLARE_TOKEN}
        else:
            return {"X-Auth-Email": self.CLOUDFLARE_EMAIL, "X-Auth-Key": self.CLOUDFLARE_API_KEY}



import time
import copy
import json
from hashlib import sha256
import binascii
import platform
from typing import Dict, Union, cast

import requests
import OpenSSL
import cryptography

from . import __version__ as sewer_version
from .config import ACME_DIRECTORY_URL_PRODUCTION
from .lib import create_logger, log_response, safe_base64


class Client:
    """
    todo: improve documentation.

    usage:
        import sewer
        provider = sewer.CloudFlareDns(
            CLOUDFLARE_EMAIL='example@example.com',
            CLOUDFLARE_API_KEY='nsa-grade-api-key'
        )

        ### to create a new certificate.
        client = sewer.Client(
            domain_name='example.com', provider=provider
        )
        certificate = client.cert()
        certificate_key = client.certificate_key
        account_key = client.account_key

        with open('certificate.crt', 'w') as certificate_file:
            certificate_file.write(certificate)

        with open('certificate.key', 'w') as certificate_key_file:
            certificate_key_file.write(certificate_key)

        ### to renew a certificate:
        with open('account_key.key', 'r') as account_key_file:
            account_key = account_key_file.read()

        client = sewer.Client(
            domain_name='example.com', provider=provider,
            account_key=account_key
        )
        certificate = client.renew()
        certificate_key = client.certificate_key

    todo: handle more exceptions
    """

    def __init__(
        self,
        domain_name,
        dns_class=None,
        domain_alt_names=None,
        contact_email=None,
        account_key=None,
        certificate_key=None,
        bits=2048,
        digest="sha256",
        provider=None,
        ACME_REQUEST_TIMEOUT=7,
        ACME_AUTH_STATUS_WAIT_PERIOD=8,
        ACME_AUTH_STATUS_MAX_CHECKS=3,
        ACME_DIRECTORY_URL=ACME_DIRECTORY_URL_PRODUCTION,
        ACME_VERIFY=True,
        LOG_LEVEL="INFO",
    ):
        """
        :param domain_name:                  (required) [string]
            the name that you want to acquire/renew certificate for. wildcards are allowed.
        :param dns_class:                    (required) [class]
            (DEPRECATED) a subclass of BaseDns which will be called to create/delete DNS TXT records.
            do not pass this parameter if also passing provider.
        :param provider:                (required) [class]
            a subclass of ProviderBase which will be called to create/delete auth records.
            do not pass this parameter if also passing dns_class
        :param domain_alt_names:             (optional) [list]
            list of alternative names that you want to be bundled into the same certificate as domain_name.
        :param contact_email:                (optional) [string]
            a contact email address
        :param account_key:                  (optional) [string]
            a string whose contents is an ssl certificate that identifies your account on the acme server.
            if you do not provide one, this client will issue a new certificate else will renew.
        :param certificate_key:              (optional) [string]
            a string whose contents is a private key that will be incorporated into your new certificate.
            if you do not provide one, this client will issue a new certificate else will renew.
        :param bits:                         (optional) [integer]
            number of bits that will be used to create your certificates' private key.
        :param digest:                       (optional) [string]
            the ssl digest type to be used in signing the certificate signing request(csr)
        :param ACME_REQUEST_TIMEOUT:         (optional) [integer]
            the max time that the client will wait for a network call to complete.
        :param ACME_AUTH_STATUS_WAIT_PERIOD: (optional) [integer]
            the interval between two consecutive client polls on the acme server to check on authorization status
        :param ACME_AUTH_STATUS_MAX_CHECKS:  (optional) [integer]
            the max number of times the client will poll the acme server to check on authorization status
        :param ACME_DIRECTORY_URL:           (optional) [string]
            the url of the acme servers' directory endpoint
        :param ACME_VERIFY:                  (optional) [bool]
            suppress verification of SSL cert when set to False (for pebble); hint: -Wignore
        :param LOG_LEVEL:                    (optional) [string]
            the level to output log messages at. one of; 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'
        """

        if not isinstance(domain_alt_names, (type(None), list)):
            raise ValueError(
                """domain_alt_names should be of type:: None or list. You entered {0}""".format(
                    type(domain_alt_names)
                )
            )
        elif not isinstance(contact_email, (type(None), str)):
            raise ValueError(
                """contact_email should be of type:: None or str. You entered {0}""".format(
                    type(contact_email)
                )
            )
        elif not isinstance(account_key, (type(None), str)):
            raise ValueError(
                """account_key should be of type:: None or str. You entered {0}.
                More specifically, account_key should be the result of reading an ssl account certificate""".format(
                    type(account_key)
                )
            )
        elif not isinstance(certificate_key, (type(None), str)):
            raise ValueError(
                """certificate_key should be of type:: None or str. You entered {0}.
                More specifically, certificate_key should be the result of reading an ssl certificate""".format(
                    type(certificate_key)
                )
            )
        elif LOG_LEVEL.upper() not in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
            raise ValueError(
                """LOG_LEVEL should be one of; 'DEBUG', 'INFO', 'WARNING', 'ERROR' or 'CRITICAL'. not {0}""".format(
                    LOG_LEVEL
                )
            )
        elif dns_class is not None and provider is not None:
            raise ValueError(
                "passed both the DEPRECATED 'dns_class' parameter as well as 'provider'."
            )

        self.domain_name = domain_name
        self.provider = provider if provider is not None else dns_class
        if not domain_alt_names:
            domain_alt_names = []
        self.domain_alt_names = domain_alt_names
        self.domain_alt_names = list(set(self.domain_alt_names))
        self.contact_email = contact_email
        self.bits = bits
        self.digest = digest
        self.ACME_REQUEST_TIMEOUT = ACME_REQUEST_TIMEOUT
        self.ACME_AUTH_STATUS_WAIT_PERIOD = ACME_AUTH_STATUS_WAIT_PERIOD
        self.ACME_AUTH_STATUS_MAX_CHECKS = ACME_AUTH_STATUS_MAX_CHECKS
        self.ACME_DIRECTORY_URL = ACME_DIRECTORY_URL
        self.ACME_VERIFY = ACME_VERIFY
        self.LOG_LEVEL = LOG_LEVEL.upper()

        self.logger = create_logger(__name__, LOG_LEVEL)

        try:
            self.all_domain_names = copy.copy(self.domain_alt_names)
            self.all_domain_names.insert(0, self.domain_name)
            self.domain_alt_names = list(set(self.domain_alt_names))

            self.User_Agent = self.get_user_agent()
            acme_endpoints = self.get_acme_endpoints().json()
            self.ACME_GET_NONCE_URL = acme_endpoints["newNonce"]
            self.ACME_TOS_URL = acme_endpoints["meta"]["termsOfService"]
            self.ACME_KEY_CHANGE_URL = acme_endpoints["keyChange"]
            self.ACME_NEW_ACCOUNT_URL = acme_endpoints["newAccount"]
            self.ACME_NEW_ORDER_URL = acme_endpoints["newOrder"]
            self.ACME_REVOKE_CERT_URL = acme_endpoints["revokeCert"]

            # unique account identifier
            # https://tools.ietf.org/html/draft-ietf-acme-acme#section-6.2
            self.kid = None

            self.certificate_key = certificate_key or self.create_certificate_key()
            self.csr = self.create_csr()

            if not account_key:
                self.account_key = self.create_account_key()
                self.PRIOR_REGISTERED = False
            else:
                self.account_key = account_key
                self.PRIOR_REGISTERED = True

            if dns_class is not None:
                self.logger.warning(
                    "DEPRECATED parameter 'dns_class' will be removed in 0.9; use 'provider' instead"
                )

            self.logger.info(
                "intialise_success, sewer_version={0}, domain_names={1}, acme_server={2}".format(
                    sewer_version.__version__,
                    self.all_domain_names,
                    self.ACME_DIRECTORY_URL[:20] + "...",
                )
            )

        ### FIX ME ### [:100] is bandaid to reduce spew during tests

        except Exception as e:
            self.logger.error("Unable to intialise Client. error={0}".format(str(e)[:100]))
            raise e

    def GET(self, url: str, **kwargs: Union[str, Dict[str, str], bool]) -> requests.Response:
        """
        wrap requests.get (and post and head, below) to allow:
          * injection of e.g. UserAgent header in one place rather than all over
          * hides requests itself to allow for change (unlikely) or use of Session
          * paves the way to inject the verify option, required to use pebble
        """

        return self._request("GET", url, **kwargs)

    def HEAD(self, url: str, **kwargs: Union[str, Dict[str, str], bool]) -> requests.Response:
        return self._request("HEAD", url, **kwargs)

    def POST(self, url: str, **kwargs: Union[str, Dict[str, str], bool]) -> requests.Response:
        return self._request("POST", url, **kwargs)

    def _request(
        self, method: str, url: str, **kwargs: Union[str, Dict[str, str], bool]
    ) -> requests.Response:
        """
        shared implementation for GET, POST and HEAD
        * injects standard request options unless they are already given in kwargs
          * header:UserAgent, timeout
          * verify - this is a hack to make sewer accept pebble's intentionally bogus cert
        """

        # if there's no UserAgent header in args, inject it
        headers = cast(Dict[str, str], kwargs.setdefault("headers", {}))
        if "UserAgent" not in headers:
            headers["UserAgent"] = self.User_Agent

        # add standard timeout if there's none already present
        if "timeout" not in kwargs:
            kwargs["timeout"] = self.ACME_REQUEST_TIMEOUT

        ### FIX ME ### can get current bogus cert from pebble, figure out how to use it here?

        # if ACME_VERIFY is false, disable certificate check in request
        if not self.ACME_VERIFY:
            kwargs["verify"] = False

        # awkward implementation to maintain compatibility with current mocked tests
        if method == "GET":
            ### Incredibly stupid hack to reconcile mypy to requests' API
            response = requests.get(url, None, **kwargs)
        elif method == "HEAD":
            response = requests.head(url, **kwargs)
        elif method == "POST":
            ### Incredibly stupid hack to reconcile mypy to requests' API
            data_hack = cast(Union[Dict[str, str], None], kwargs.pop("data", None))
            response = requests.post(url, data_hack, **kwargs)

        return response

    @staticmethod
    def get_user_agent():
        return "python-requests/{requests_version} ({system}: {machine}) sewer {sewer_version} ({sewer_url})".format(
            requests_version=requests.__version__,
            system=platform.system(),
            machine=platform.machine(),
            sewer_version=sewer_version.__version__,
            sewer_url=sewer_version.__url__,
        )

    def get_acme_endpoints(self):
        self.logger.debug("get_acme_endpoints")
        get_acme_endpoints = self.GET(self.ACME_DIRECTORY_URL)
        self.logger.debug(
            "get_acme_endpoints_response. status_code={0}".format(get_acme_endpoints.status_code)
        )
        if get_acme_endpoints.status_code not in [200, 201]:
            raise ValueError(
                "Error while getting Acme endpoints: status_code={status_code} response={response}".format(
                    status_code=get_acme_endpoints.status_code,
                    response=log_response(get_acme_endpoints),
                )
            )
        return get_acme_endpoints

    def create_certificate_key(self):
        self.logger.debug("create_certificate_key")
        return self.create_key().decode()

    def create_account_key(self):
        self.logger.debug("create_account_key")
        return self.create_key().decode()

    def create_key(self, key_type=OpenSSL.crypto.TYPE_RSA):
        key = OpenSSL.crypto.PKey()
        key.generate_key(key_type, self.bits)
        private_key = OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM, key)
        return private_key

    def create_csr(self):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.4
        The CSR is sent in the base64url-encoded version of the DER format. (NB: this
        field uses base64url, and does not include headers, it is different from PEM.)
        """
        self.logger.debug("create_csr")
        X509Req = OpenSSL.crypto.X509Req()
        X509Req.get_subject().CN = self.domain_name

        if self.domain_alt_names:
            SAN = "DNS:{0}, ".format(self.domain_name).encode("utf8") + ", ".join(
                "DNS:" + i for i in self.domain_alt_names
            ).encode("utf8")
        else:
            SAN = "DNS:{0}".format(self.domain_name).encode("utf8")

        X509Req.add_extensions(
            [
                OpenSSL.crypto.X509Extension(
                    "subjectAltName".encode("utf8"), critical=False, value=SAN
                )
            ]
        )
        pk = OpenSSL.crypto.load_privatekey(
            OpenSSL.crypto.FILETYPE_PEM, self.certificate_key.encode()
        )
        X509Req.set_pubkey(pk)
        X509Req.set_version(2)
        X509Req.sign(pk, self.digest)
        return OpenSSL.crypto.dump_certificate_request(OpenSSL.crypto.FILETYPE_ASN1, X509Req)

    def acme_register(self):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.3
        The server creates an account and stores the public key used to
        verify the JWS (i.e., the "jwk" element of the JWS header) to
        authenticate future requests from the account.
        The server returns this account object in a 201 (Created) response, with the account URL
        in a Location header field.
        This account URL will be used in subsequest requests to ACME, as the "kid" value in the acme header.
        If the server already has an account registered with the provided
        account key, then it MUST return a response with a 200 (OK) status
        code and provide the URL of that account in the Location header field.
        If there is an existing account with the new key
        provided, then the server SHOULD use status code 409 (Conflict) and
        provide the URL of that account in the Location header field
        """
        self.logger.info("acme_register (newAccount)")
        if self.PRIOR_REGISTERED:
            payload = {"onlyReturnExisting": True}
        elif self.contact_email:
            payload = {
                "termsOfServiceAgreed": True,
                "contact": ["mailto:{0}".format(self.contact_email)],
            }
        else:
            payload = {"termsOfServiceAgreed": True}

        url = self.ACME_NEW_ACCOUNT_URL
        acme_register_response = self.make_signed_acme_request(
            url=url, payload=json.dumps(payload), needs_jwk=True
        )
        self.logger.debug(
            "acme_register_response. status_code={0}. response={1}".format(
                acme_register_response.status_code, log_response(acme_register_response)
            )
        )

        if acme_register_response.status_code not in [201, 200, 409]:
            raise ValueError(
                "Error while registering: status_code={status_code} response={response}".format(
                    status_code=acme_register_response.status_code,
                    response=log_response(acme_register_response),
                )
            )

        kid = acme_register_response.headers["Location"]
        setattr(self, "kid", kid)

        self.logger.info("acme_register_success")
        return acme_register_response

    def apply_for_cert_issuance(self):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.4
        The order object returned by the server represents a promise that if
        the client fulfills the server's requirements before the "expires"
        time, then the server will be willing to finalize the order upon
        request and issue the requested certificate.  In the order object,
        any authorization referenced in the "authorizations" array whose
        status is "pending" represents an authorization transaction that the
        client must complete before the server will issue the certificate.

        Once the client believes it has fulfilled the server's requirements,
        it should send a POST request to the order resource's finalize URL.
        The POST body MUST include a CSR:

        The date values seem to be ignored by LetsEncrypt although they are
        in the ACME draft spec; https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.4
        """
        self.logger.info("apply_for_cert_issuance (newOrder)")
        identifiers = []
        for domain_name in self.all_domain_names:
            identifiers.append({"type": "dns", "value": domain_name})

        payload = {"identifiers": identifiers}
        url = self.ACME_NEW_ORDER_URL
        apply_for_cert_issuance_response = self.make_signed_acme_request(
            url=url, payload=json.dumps(payload)
        )
        self.logger.debug(
            "apply_for_cert_issuance_response. status_code={0}. response={1}".format(
                apply_for_cert_issuance_response.status_code,
                log_response(apply_for_cert_issuance_response),
            )
        )

        if apply_for_cert_issuance_response.status_code != 201:
            raise ValueError(
                "Error applying for certificate issuance: status_code={status_code} response={response}".format(
                    status_code=apply_for_cert_issuance_response.status_code,
                    response=log_response(apply_for_cert_issuance_response),
                )
            )

        apply_for_cert_issuance_response_json = apply_for_cert_issuance_response.json()
        finalize_url = apply_for_cert_issuance_response_json["finalize"]
        authorizations = apply_for_cert_issuance_response_json["authorizations"]

        self.logger.info("apply_for_cert_issuance_success")
        return authorizations, finalize_url

    def get_identifier_authorization(self, auth_url: str) -> Dict[str, str]:
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.5
        When a client receives an order from the server it downloads the
        authorization resources by sending GET requests to the indicated
        URLs.  If the client initiates authorization using a request to the
        new authorization resource, it will have already received the pending
        authorization object in the response to that request.

        This is also where we get the challenges/tokens.
        """
        self.logger.info("get_identifier_authorization for %s" % auth_url)
        response = self.make_signed_acme_request(auth_url, payload="")

        self.logger.debug(
            "get_identifier_authorization_response. status_code={0}. response={1}".format(
                response.status_code, log_response(response)
            )
        )
        if response.status_code not in [200, 201]:
            raise ValueError(
                "Error getting identifier authorization: status_code={status_code} response={response}".format(
                    status_code=response.status_code, response=log_response(response)
                )
            )
        response_json = response.json()
        domain = response_json["identifier"]["value"]
        wildcard = response_json.get("wildcard")

        for i in response_json["challenges"]:
            if i["type"] in self.provider.chal_types:
                challenge = i
                challenge_token = challenge["token"]
                challenge_url = challenge["url"]

                identifier_auth = {
                    "domain": domain,
                    "url": auth_url,
                    "wildcard": wildcard,
                    "token": challenge_token,
                    "challenge_url": challenge_url,
                }

        self.logger.debug(
            "get_identifier_authorization_success. identifier_auth={0}".format(identifier_auth)
        )
        self.logger.info(
            "get_identifier_authorization got %s, token=%s" % (challenge_url, challenge_token)
        )
        return identifier_auth

    def get_keyauthorization(self, token):
        self.logger.debug("get_keyauthorization")
        acme_header_jwk_json = json.dumps(self.get_jwk(), sort_keys=True, separators=(",", ":"))
        acme_thumbprint = safe_base64(sha256(acme_header_jwk_json.encode("utf8")).digest())
        acme_keyauthorization = "{0}.{1}".format(token, acme_thumbprint)

        return acme_keyauthorization

    def check_authorization_status(self, authorization_url, desired_status=None):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.5.1
        To check on the status of an authorization, the client sends a GET(polling)
        request to the authorization URL, and the server responds with the
        current authorization object.

        https://tools.ietf.org/html/draft-ietf-acme-acme#section-8.2
        Clients SHOULD NOT respond to challenges until they believe that the
        server's queries will succeed. If a server's initial validation
        query fails, the server SHOULD retry[intended to address things like propagation delays in
        HTTP/DNS provisioning] the query after some time.
        The server MUST provide information about its retry state to the
        client via the "errors" field in the challenge and the Retry-After
        """
        self.logger.info("check_authorization_status")
        desired_status = desired_status or ["pending", "valid"]
        number_of_checks = 0
        while True:
            time.sleep(self.ACME_AUTH_STATUS_WAIT_PERIOD)
            check_authorization_status_response = self.make_signed_acme_request(
                authorization_url, payload=""
            )
            authorization_status = check_authorization_status_response.json()["status"]
            number_of_checks = number_of_checks + 1
            self.logger.debug(
                "check_authorization_status_response. status_code={0}. response={1}".format(
                    check_authorization_status_response.status_code,
                    log_response(check_authorization_status_response),
                )
            )
            if authorization_status in desired_status:
                break
            if number_of_checks == self.ACME_AUTH_STATUS_MAX_CHECKS:
                raise StopIteration(
                    "Checks done={0}. Max checks allowed={1}. Interval between checks={2}seconds.".format(
                        number_of_checks,
                        self.ACME_AUTH_STATUS_MAX_CHECKS,
                        self.ACME_AUTH_STATUS_WAIT_PERIOD,
                    )
                )

        self.logger.info("check_authorization_status_success")
        return check_authorization_status_response

    def respond_to_challenge(self, acme_keyauthorization, challenge_url):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.5.1
        To prove control of the identifier and receive authorization, the
        client needs to respond with information to complete the challenges.
        The server is said to "finalize" the authorization when it has
        completed one of the validations, by assigning the authorization a
        status of "valid" or "invalid".

        Usually, the validation process will take some time, so the client
        will need to poll the authorization resource to see when it is finalized.
        To check on the status of an authorization, the client sends a GET(polling)
        request to the authorization URL, and the server responds with the
        current authorization object.
        """
        self.logger.info(
            "respond_to_challenge for %s at %s" % (acme_keyauthorization, challenge_url)
        )
        payload = json.dumps({"keyAuthorization": "{0}".format(acme_keyauthorization)})
        respond_to_challenge_response = self.make_signed_acme_request(challenge_url, payload)
        self.logger.debug(
            "respond_to_challenge_response. status_code={0}. response={1}".format(
                respond_to_challenge_response.status_code,
                log_response(respond_to_challenge_response),
            )
        )

        self.logger.info("respond_to_challenge_success")
        return respond_to_challenge_response

    def send_csr(self, finalize_url):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-7.4
        Once the client believes it has fulfilled the server's requirements,
        it should send a POST request(include a CSR) to the order resource's finalize URL.
        A request to finalize an order will result in error if the order indicated does not have status "pending",
        if the CSR and order identifiers differ, or if the account is not authorized for the identifiers indicated in the CSR.
        The CSR is sent in the base64url-encoded version of the DER format(OpenSSL.crypto.FILETYPE_ASN1)

        A valid request to finalize an order will return the order to be finalized.
        The client should begin polling the order by sending a
        GET request to the order resource to obtain its current state.
        """
        self.logger.info("send_csr")
        payload = {"csr": safe_base64(self.csr)}
        send_csr_response = self.make_signed_acme_request(
            url=finalize_url, payload=json.dumps(payload)
        )
        self.logger.debug(
            "send_csr_response. status_code={0}. response={1}".format(
                send_csr_response.status_code, log_response(send_csr_response)
            )
        )

        if send_csr_response.status_code not in [200, 201]:
            raise ValueError(
                "Error sending csr: status_code={status_code} response={response}".format(
                    status_code=send_csr_response.status_code,
                    response=log_response(send_csr_response),
                )
            )
        send_csr_response_json = send_csr_response.json()
        certificate_url = send_csr_response_json["certificate"]

        self.logger.info("send_csr_success")
        return certificate_url

    def download_certificate(self, certificate_url: str) -> str:
        self.logger.info("download_certificate")

        response = self.make_signed_acme_request(certificate_url, payload="")
        self.logger.debug(
            "download_certificate_response. status_code={0}. response={1}".format(
                response.status_code, log_response(response)
            )
        )
        if response.status_code not in [200, 201]:
            raise ValueError(
                "Error fetching signed certificate: status_code={status_code} response={response}".format(
                    status_code=response.status_code, response=log_response(response)
                )
            )
        pem_certificate = response.content.decode("utf-8")
        self.logger.info("download_certificate_success")
        return pem_certificate

    def sign_message(self, message):
        self.logger.debug("sign_message")
        pk = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, self.account_key.encode())
        return OpenSSL.crypto.sign(pk, message.encode("utf8"), self.digest)

    def get_nonce(self):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-6.4
        Each request to an ACME server must include a fresh unused nonce
        in order to protect against replay attacks.
        """
        self.logger.debug("get_nonce")
        response = self.GET(self.ACME_GET_NONCE_URL)
        nonce = response.headers["Replay-Nonce"]
        return nonce

    @staticmethod
    def stringfy_items(payload):
        """
        method that takes a dictionary and then converts any keys or values
        in that are of type bytes into unicode strings.
        This is necessary esp if you want to then turn that dict into a json string.
        """
        if isinstance(payload, str):
            return payload

        for k, v in payload.items():
            if isinstance(k, bytes):
                k = k.decode("utf-8")
            if isinstance(v, bytes):
                v = v.decode("utf-8")
            payload[k] = v
        return payload

    def get_jwk(self):
        """
        calculate the JSON Web Key (jwk) from self.account_key
        """
        private_key = cryptography.hazmat.primitives.serialization.load_pem_private_key(
            self.account_key.encode(),
            password=None,
            backend=cryptography.hazmat.backends.default_backend(),
        )
        public_key_public_numbers = private_key.public_key().public_numbers()
        # private key public exponent in hex format
        exponent = "{0:x}".format(public_key_public_numbers.e)
        exponent = "0{0}".format(exponent) if len(exponent) % 2 else exponent
        # private key modulus in hex format
        modulus = "{0:x}".format(public_key_public_numbers.n)
        jwk = {
            "kty": "RSA",
            "e": safe_base64(binascii.unhexlify(exponent)),
            "n": safe_base64(binascii.unhexlify(modulus)),
        }
        return jwk

    def get_acme_header(self, url, needs_jwk=False):
        """
        https://tools.ietf.org/html/draft-ietf-acme-acme#section-6.2
        The JWS Protected Header MUST include the following fields:
        - "alg" (Algorithm)
        - "jwk" (JSON Web Key, only for requests to new-account and revoke-cert resources)
        - "kid" (Key ID, for all other requests). gotten from self.ACME_NEW_ACCOUNT_URL
        - "nonce". gotten from self.ACME_GET_NONCE_URL
        - "url"
        """
        self.logger.debug("get_acme_header")
        header = {"alg": "RS256", "nonce": self.get_nonce(), "url": url}

        if needs_jwk:
            header["jwk"] = self.get_jwk()
        else:
            header["kid"] = self.kid
        return header

    def make_signed_acme_request(self, url, payload, needs_jwk=False):
        self.logger.debug("make_signed_acme_request")
        headers = {}
        payload64 = safe_base64(payload)
        protected = self.get_acme_header(url, needs_jwk)
        protected64 = safe_base64(json.dumps(protected))
        signature = self.sign_message(message="{0}.{1}".format(protected64, payload64))  # bytes
        signature64 = safe_base64(signature)  # str
        data = json.dumps(
            {"protected": protected64, "payload": payload64, "signature": signature64}
        )
        headers.update({"Content-Type": "application/jose+json"})
        response = self.POST(url, data=data.encode("utf8"), headers=headers)
        return response

    def get_certificate(self):
        self.logger.debug("get_certificate")
        challenges = []

        try:
            self.acme_register()
            authorizations, finalize_url = self.apply_for_cert_issuance()

            for auth_url in authorizations:
                identifier_auth = self.get_identifier_authorization(auth_url)
                token = identifier_auth["token"]
                challenge = {
                    "ident_value": identifier_auth["domain"],
                    "token": token,
                    "key_auth": self.get_keyauthorization(token),  # responder acme_keyauth..
                    "wildcard": identifier_auth["wildcard"],
                    "auth_url": auth_url,  # responder auth.._url
                    "chal_url": identifier_auth["challenge_url"],  # responder challenge_url
                }
                challenges.append(challenge)

            # any errors in setup are fatal (here - they are all necessary for same cert)
            failures = self.provider.setup(challenges)
            if failures:
                raise RuntimeError("get_certificate: challenge setup failed for %s" % failures)

            ### FIX ME ### should abort cert and try to clear on error

            self.propagation_delay(challenges)

            # for a case where you want certificates for *.example.com and example.com
            # you have to create both auth records AND then respond to the challenge.
            # see issues/83
            for chal in challenges:
                # Make sure the authorization is in a status where we can submit a challenge
                # response. The authorization can be in the "valid" state before submitting
                # a challenge response if there was a previous authorization for these hosts
                # that was successfully validated, still cached by the server.
                auth_status_response = self.check_authorization_status(chal["auth_url"])
                if auth_status_response.json()["status"] == "pending":
                    self.respond_to_challenge(chal["key_auth"], chal["chal_url"])

            ### TO DO ### this is the obfuscated timeout loop.  Clean this mess up!
            ### # # # ### it also keeps trying even when the auth is failed :-(

            for chal in challenges:
                # Before sending a CSR, we need to make sure the server has completed the
                # validation for all the authorizations
                self.check_authorization_status(chal["auth_url"], ["valid"])

            certificate_url = self.send_csr(finalize_url)
            certificate = self.download_certificate(certificate_url)

        ### FIX ME ### [:100] is a bandaid to reduce spew during tests

        except Exception as e:
            self.logger.error("Error: Unable to issue certificate. error={0}".format(str(e)[:100]))
            raise e
        finally:
            # best-effort attempt to clear challenges
            failures = self.provider.clear(challenges)

        return certificate

    def sleep_iter(self):
        "returns values from list, then repeats last value forever"

        for cur_time in self.provider.prop_sleep_times:
            yield cur_time
        while True:
            yield cur_time

    def propagation_delay(self, challenges):
        """
        Wait for the challenges to propagate through the service.

        SHOULD return success or failure... TBD.

        This process is controlled by two attributes and one method on the provider.
        unpropagated() is called to check whether the listed challenges are thought
        to be ready to be validated by the ACME server.  If it fails, propagation_delay
        will sleep and retry until it times out.  The attributes on the provider are
        prop_timeout and prop_sleep_times.  See docs/unpropagated.md for the details.
        """

        if self.provider.prop_timeout:
            unready = challenges
            end_time = time.time() + self.provider.prop_timeout
            next_sleep = self.sleep_iter()
            num_checks = 0
            while unready:
                errata = self.provider.unpropagated(unready)
                num_checks += 1

                ### FIX ME ### if any errata have status "failed", exit with error
                if errata:
                    poll_time = time.time()
                    # intentional: do an "extra" check rather than running short
                    if end_time < poll_time:
                        break
                    # wait a while to let more propagation happen
                    time.sleep(next(next_sleep))

                unready = [err[2] for err in errata]

            ### FIX ME ### might be good for mock tests, but really should try to clear, eh?
            if unready:
                raise RuntimeError(
                    "propagation_delay: time out after %s probes: %s" % (num_checks, unready)
                )

    def cert(self):
        """
        convenience method to get a certificate without much hassle
        """
        return self.get_certificate()

    def renew(self):
        """
        renews a certificate.
        A renewal is actually just getting a new certificate.
        An issuance request counts as a renewal if it contains the exact same set of hostnames as a previously issued certificate.
            https://letsencrypt.org/docs/rate-limits/
        """
        return self.cert()



import os

from setuptools import setup, find_packages

# To use a consistent encoding
import codecs

here = os.path.abspath(os.path.dirname(__file__))
about = {}

try:
    import pypandoc

    long_description = pypandoc.convert("README.md", "rst")
except ImportError:
    long_description = codecs.open("README.md", encoding="utf8").read()

with open(os.path.join(here, "sewer", "__version__.py"), "r") as f:
    exec(f.read(), about)

dns_provider_deps_map = {
    "cloudflare": [""],
    "aliyun": ["aliyun-python-sdk-core-v3", "aliyun-python-sdk-alidns"],
    "hurricane": ["hurricanedns"],
    "aurora": ["tldextract", "apache-libcloud"],
    "acmedns": ["dnspython"],
    "rackspace": ["tldextract"],
    "dnspod": [""],
    "duckdns": [""],
    "cloudns": ["cloudns-api"],
    "route53": ["boto3"],
    "powerdns": [""],
}

all_deps_of_all_dns_provider = []
for _, vlist in dns_provider_deps_map.items():
    all_deps_of_all_dns_provider += vlist
all_deps_of_all_dns_provider = list(set(all_deps_of_all_dns_provider))

setup(
    name=about["__title__"],
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=about["__version__"],
    description=about["__description__"],
    long_description=long_description,
    # The project's main homepage.
    url=about["__url__"],
    # Author details
    author=about["__author__"],
    author_email=about["__author_email__"],
    # Choose your license
    license=about["__license__"],
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Security",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Networking",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    # What does your project relate to?
    keywords="letsencrypt",
    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    # packages=['sewer'],
    packages=find_packages(exclude=["docs", "*tests*"]),
    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["requests", "pyopenssl", "cryptography"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip3 install -e .[dev,test]
    extras_require={
        "dev": ["coverage", "pypandoc", "twine", "wheel"],
        "test": ["pylint==2.3.1", "black==18.9b0"],
        "cloudflare": dns_provider_deps_map["cloudflare"],
        "aliyun": dns_provider_deps_map["aliyun"],
        "hurricane": dns_provider_deps_map["hurricane"],
        "aurora": dns_provider_deps_map["aurora"],
        "acmedns": dns_provider_deps_map["acmedns"],
        "rackspace": dns_provider_deps_map["rackspace"],
        "dnspod": dns_provider_deps_map["dnspod"],
        "duckdns": dns_provider_deps_map["duckdns"],
        "cloudns": dns_provider_deps_map["cloudns"],
        "route53": dns_provider_deps_map["route53"],
        "powerdns": dns_provider_deps_map["powerdns"],
        "alldns": all_deps_of_all_dns_provider,
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },
    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
    entry_points={"console_scripts": ["sewer=sewer.cli:main", "sewer-cli=sewer.cli:main"]},
)

# python packaging documentation:
# 1. https://python-packaging.readthedocs.io/en/latest/index.html
# 2. https://python-packaging-user-guide.readthedocs.io/tutorials/distributing-packages
# a) pip3 install wheel twine
# b) pip3 install -e .
# c) python setup.py sdist
# d) python setup.py bdist_wheel
# e) DONT use python setup.py register and python setup.py upload. They use http
# f) twine upload dist/* -r testpypi
# g) pip3 install -i https://testpypi.python.org/pypi <package name>
# h) twine upload dist/*   # prod pypi
# i) pip3 install <package name>
